from __future__ import annotations

import inspect
import logging
import typing
from collections import Counter, defaultdict
from typing import Any, Callable, Collection, Dict, List, Optional, Tuple, Type, Union

import pandas as pd

from hamilton import models, node
from hamilton.dev_utils.deprecation import deprecated
from hamilton.function_modifiers import base
from hamilton.function_modifiers.configuration import ConfigResolver, hamilton_exclude
from hamilton.function_modifiers.delayed import resolve as delayed_resolve
from hamilton.function_modifiers.dependencies import (
    LiteralDependency,
    SingleDependency,
    UpstreamDependency,
    source,
)

logger = logging.getLogger(__name__)

"""Decorators that replace a function's execution with specified behavior"""

# Python 3.10 + has this built in, otherwise we have to define it
try:
    from types import EllipsisType
except ImportError:
    EllipsisType = type(...)


# the following are empty functions that we can compare against to ensure that @does uses an empty function
def _empty_function():
    pass


def _empty_function_with_docstring():
    """Docstring for an empty function"""
    pass


def ensure_function_empty(fn: Callable):
    """
    Ensures that a function is empty. This is strict definition -- the function must have only one line (and
    possibly a docstring), and that line must say "pass".
    """
    if fn.__code__.co_code not in {
        _empty_function.__code__.co_code,
        _empty_function_with_docstring.__code__.co_code,
    }:
        raise base.InvalidDecoratorException(
            f"Function: {fn.__name__} is not empty. Must have only one line that "
            'consists of "pass"'
        )


class does(base.NodeCreator):
    """``@does`` is a decorator that essentially allows you to run a function over all the input parameters. \
    So you can't pass any old function to ``@does``, instead the function passed has to take any amount of inputs and \
    process them all in the same way.

    .. code-block:: python

        import pandas as pd
        from hamilton.function_modifiers import does
        import internal_package_with_logic

        def sum_series(**series: pd.Series) -> pd.Series:
            '''This function takes any number of inputs and sums them all together.'''
            ...

        @does(sum_series)
        def D_XMAS_GC_WEIGHTED_BY_DAY(D_XMAS_GC_WEIGHTED_BY_DAY_1: pd.Series,
                                      D_XMAS_GC_WEIGHTED_BY_DAY_2: pd.Series) -> pd.Series:
            '''Adds D_XMAS_GC_WEIGHTED_BY_DAY_1 and D_XMAS_GC_WEIGHTED_BY_DAY_2'''
            pass

        @does(internal_package_with_logic.identity_function)
        def copy_of_x(x: pd.Series) -> pd.Series:
            '''Just returns x'''
            pass

    The example here is a function, that all that it does, is sum all the parameters together. So we can annotate it \
    with the ``@does`` decorator and pass it the ``sum_series`` function. The ``@does`` decorator is currently limited \
    to just allow functions that consist only of one argument, a generic \\*\\*kwargs.
    """

    def __init__(self, replacing_function: Callable, **argument_mapping: Union[str, List[str]]):
        """Constructor for a modifier that replaces the annotated functions functionality with something else.
        Right now this has a very strict validation requirements to make compliance with the framework easy.

        :param replacing_function: The function to replace the original function with.
        :param argument_mapping: A mapping of argument name in the replacing function to argument name in the \
        decorating function.
        """
        self.replacing_function = replacing_function
        self.argument_mapping = argument_mapping

    @staticmethod
    def map_kwargs(kwargs: Dict[str, Any], argument_mapping: Dict[str, str]) -> Dict[str, Any]:
        """Maps kwargs using the argument mapping.
        This does 2 things:
        1. Replaces all kwargs in passed_in_kwargs with their mapping
        2. Injects all defaults from the origin function signature

        :param kwargs: Keyword arguments that will be passed into a hamilton function.
        :param argument_mapping: Mapping of those arguments to a replacing function's arguments.
        :return: The new kwargs for the replacing function's arguments.
        """
        output = {**kwargs}
        for arg_mapped_to, original_arg in argument_mapping.items():
            if original_arg in kwargs and arg_mapped_to not in argument_mapping.values():
                del output[original_arg]
            # Note that if it is not there it could be a **kwarg
            output[arg_mapped_to] = kwargs[original_arg]
        return output

    @staticmethod
    def test_function_signatures_compatible(
        fn_signature: inspect.Signature,
        replace_with_signature: inspect.Signature,
        argument_mapping: Dict[str, str],
    ) -> bool:
        """Tests whether a function signature and the signature of the replacing function are compatible.

        :param fn_signature:
        :param replace_with_signature:
        :param argument_mapping:
        :return: True if they're compatible, False otherwise
        """
        # The easy (and robust) way to do this is to use the bind with a set of dummy arguments and test if it breaks.
        # This way we're not reinventing the wheel.
        SENTINEL_ARG_VALUE = ...  # does not matter as we never use it
        # We initialize as the default values, as they'll always be injected in
        dummy_param_values = {
            key: SENTINEL_ARG_VALUE
            for key, param_spec in fn_signature.parameters.items()
            if param_spec.default is not inspect.Parameter.empty
        }
        # Then we update with the dummy values. Again, replacing doesn't matter (we'll be mimicking it later)
        dummy_param_values.update({key: SENTINEL_ARG_VALUE for key in fn_signature.parameters})
        dummy_param_values = does.map_kwargs(dummy_param_values, argument_mapping)
        try:
            # Python signatures have a bind() capability which does exactly what we want to do
            # Throws a type error if it is not valid
            replace_with_signature.bind(**dummy_param_values)
        except TypeError:
            return False
        return True

    @staticmethod
    def ensure_function_signature_compatible(
        og_function: Callable,
        replacing_function: Callable,
        argument_mapping: Dict[str, str],
    ):
        """Ensures that a function signature is compatible with the replacing function, given the argument mapping

        :param og_function: Function that's getting replaced (decorated with `@does`)
        :param replacing_function: A function that gets called in its place (passed in by `@does`)
        :param argument_mapping: The mapping of arguments from fn to replace_with
        :return:
        """
        fn_parameters = inspect.signature(og_function).parameters
        invalid_fn_parameters = []
        for param_name, param_spec in fn_parameters.items():
            if param_spec.kind not in {
                inspect.Parameter.KEYWORD_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            }:
                invalid_fn_parameters.append(param_name)

        if invalid_fn_parameters:
            raise base.InvalidDecoratorException(
                f"Decorated function for @does (and really, all of hamilton), "
                f"can only consist of keyword-friendly arguments. "
                f"The following parameters for {og_function.__name__} are not keyword-friendly: {invalid_fn_parameters}"
            )
        if not does.test_function_signatures_compatible(
            inspect.signature(og_function),
            inspect.signature(replacing_function),
            argument_mapping,
        ):
            raise base.InvalidDecoratorException(
                f"The following function signatures are not compatible for use with @does: "
                f"{og_function.__name__} with signature {inspect.signature(og_function)} "
                f"and replacing function {replacing_function.__name__} with signature {inspect.signature(replacing_function)}. "
                f"Mapping for arguments provided was: {argument_mapping}. You can fix this by either adjusting "
                f"the signature for the replacing function *or* adjusting the mapping."
            )

    def validate(self, fn: Callable):
        """Validates that the function:
        - Is empty (we don't want to be overwriting actual code)
        - Has a compatible return type
        - Matches the function signature with the appropriate mapping
        :param fn: Function to validate
        :raises: InvalidDecoratorException
        """
        ensure_function_empty(fn)
        does.ensure_function_signature_compatible(
            fn, self.replacing_function, self.argument_mapping
        )

    def generate_nodes(self, fn: Callable, config) -> List[node.Node]:
        """Returns one node which has the replaced functionality
        :param fn: Function to decorate
        :param config: Configuration (not used in this)
        :return: A node with the function in `@does` injected,
        and the same parameters/types as the original function.
        """

        def wrapper_function(**kwargs):
            final_kwarg_values = {
                key: param_spec.default
                for key, param_spec in inspect.signature(fn).parameters.items()
                if param_spec.default is not inspect.Parameter.empty
            }
            final_kwarg_values.update(kwargs)
            final_kwarg_values = does.map_kwargs(final_kwarg_values, self.argument_mapping)
            return self.replacing_function(**final_kwarg_values)

        return [node.Node.from_fn(fn).copy_with(callabl=wrapper_function)]


def get_default_tags(fn: Callable) -> Dict[str, str]:
    """Function that encapsulates default tags on a function.

    :param fn: the function we want to create default tags for.
    :return: a dictionary with str -> str values representing the default tags.
    """
    module_name = inspect.getmodule(fn).__name__
    return {"module": module_name}


@deprecated(
    warn_starting=(1, 20, 0),
    fail_starting=(2, 0, 0),
    use_this=delayed_resolve,
    explanation="dynamic_transform has been replaced with @resolve -- a cleaner way"
    "to utilize config for resolving decorators. Note this allows you to use any"
    "existing decorators.",
    current_version=(1, 19, 0),
    migration_guide="https://hamilton.dagworks.io/en/latest/reference/decorators/",
)
class dynamic_transform(base.NodeCreator):
    def __init__(
        self,
        transform_cls: Type[models.BaseModel],
        config_param: str,
        **extra_transform_params,
    ):
        """Constructs a model. Takes in a model_cls, which has to have a parameter."""
        self.transform_cls = transform_cls
        self.config_param = config_param
        self.extra_transform_params = extra_transform_params

    def validate(self, fn: Callable):
        """Validates that the model works with the function -- ensures:
        1. function has no code
        2. function has no parameters
        3. function has series as a return type
        :param fn: Function to validate
        :raises InvalidDecoratorException if the model is not valid.
        """

        ensure_function_empty(fn)  # it has to look exactly
        signature = inspect.signature(fn)
        if not issubclass(typing.get_type_hints(fn).get("return"), pd.Series):
            raise base.InvalidDecoratorException(
                "Models must declare their return type as a pandas Series"
            )
        if len(signature.parameters) > 0:
            raise base.InvalidDecoratorException(
                "Models must have no parameters -- all are passed in through the config"
            )

    def generate_nodes(self, fn: Callable, config: Dict[str, Any] = None) -> List[node.Node]:
        if self.config_param not in config:
            raise base.InvalidDecoratorException(
                f"Configuration has no parameter: {self.config_param}. Did you define it? If so did you spell it right?"
            )
        fn_name = fn.__name__
        transform = self.transform_cls(
            config[self.config_param], fn_name, **self.extra_transform_params
        )
        return [
            node.Node(
                name=fn_name,
                typ=typing.get_type_hints(fn).get("return"),
                doc_string=fn.__doc__,
                callabl=transform.compute,
                input_types={dep: pd.Series for dep in transform.get_dependents()},
                tags=get_default_tags(fn),
            )
        ]

    def require_config(self) -> List[str]:
        """Returns the configuration parameters that this model requires

        :return: Just the one config param used by this model
        """
        return [self.config_param]


class model(dynamic_transform):
    """Model, same as a dynamic transform"""

    def __init__(self, model_cls, config_param: str, **extra_model_params):
        super(model, self).__init__(
            transform_cls=model_cls, config_param=config_param, **extra_model_params
        )


NamespaceType = Union[str, EllipsisType, None]


class Applicable:
    """Applicable is a largely internal construct that represents a function that can be applied as a node.
    A few of these function are external-facing, however (named, when, when_not, ...)"""

    def __init__(
        self,
        fn: Union[Callable, str, None],
        args: Tuple[Union[Any, SingleDependency], ...],
        kwargs: Dict[str, Union[Any, SingleDependency]],
        target_fn: Union[Callable, str, None] = None,
        _resolvers: List[ConfigResolver] = None,
        _name: Optional[str] = None,
        _namespace: Union[str, None, EllipsisType] = ...,
        _target: base.TargetType = None,
    ):
        """Instantiates an Applicable.

        We allow fn=None for the use-cases where we want to store the Applicable config (i.e. .when* family, namespace, target, etc.)
        but do not yet the access to the actual function we are turning into the Applicable. In addition, in case the target nodes come
        from a function (using extract_columns/extract_fields) we can pass target_fn to have access to its pointer that we can decorate
        programmatically. See `apply_to` and `mutate` for an example.

        :param args: Args (*args) to pass to the function
        :param fn: Function it takes in. Can be None to create an Applicable placeholder with delayed choice of function.
        :param target_fn: Function the applicable will be applied to
        :param _resolvers: Resolvers to use for the function
        :param _name: Name of the node to be created
        :param _namespace: Namespace of the node to be created -- currently only single-level namespaces are supported
        :param _target: Selects which target nodes it will be appended onto. Default None gets resolved on decorator level.
            Specifically, pipe_input would use the first parameter and pipe_output / mutate would apply it to all sink nodes.
        :param kwargs: Kwargs (**kwargs) to pass to the function
        """

        if isinstance(fn, str) or isinstance(target_fn, str):
            raise TypeError("Strings are not supported currently. Please provide function pointer.")

        self.fn = fn
        self.target_fn = target_fn

        if "_name" in kwargs:
            raise ValueError("Cannot pass in _name as a kwarg")

        self.kwargs = {key: value for key, value in kwargs.items() if key != "__name"}  # TODO --
        self.args = args
        # figure out why this was showing up in two places...
        self.resolvers = _resolvers if _resolvers is not None else []
        self.name = _name
        self.namespace = _namespace
        self.target = _target

    def _with_resolvers(self, *additional_resolvers: ConfigResolver) -> "Applicable":
        """Helper function for the .when* group"""
        return Applicable(
            fn=self.fn,
            _resolvers=self.resolvers + list(additional_resolvers),
            _name=self.name,
            _namespace=self.namespace,
            _target=self.target,
            args=self.args,
            kwargs=self.kwargs,
            target_fn=self.target_fn,
        )

    def when(self, **key_value_pairs) -> "Applicable":
        """Choose to apply this function when all of the keys in the function
        are present in the config, and the values match the values in the config.

        :param key_value_pairs: key/value pairs to match
        :return: The Applicable with this condition applied
        """
        return self._with_resolvers(ConfigResolver.when(**key_value_pairs))

    def when_not(self, **key_value_pairs) -> "Applicable":
        """Choose to apply this function when all of the keys specified
        do not match the values specified in the config.

        :param key_value_pairs: key/value pairs to match
        :return: The Applicable with this condition applied
        """
        return self._with_resolvers(ConfigResolver.when_not(**key_value_pairs))

    def when_in(self, **key_value_group_pairs: list) -> "Applicable":
        """Choose to apply this function when all the keys provided have values contained within the list of values
        specified

        :param key_value_group_pairs: key/value pairs to match
        :return:  The Applicable with this condition applied
        """
        return self._with_resolvers(ConfigResolver.when_in(**key_value_group_pairs))

    def when_not_in(self, **key_value_group_pairs: list) -> "Applicable":
        """Choose to apply this function when all the keys provided have values not contained within the list of values
        specified.

        :param key_value_group_pairs: key/value pairs to match
        :return:  The Applicable with this condition applied
        """

        return self._with_resolvers(ConfigResolver.when_not_in(**key_value_group_pairs))

    def namespaced(self, namespace: NamespaceType) -> "Applicable":
        """Add a namespace to this node. You probably don't need this -- you should look at "named" instead.

        :param namespace: Namespace to apply, can be ..., None, or a string.
        :return: The Applicable with this namespace
        """
        return Applicable(
            fn=self.fn,
            _resolvers=self.resolvers,
            _name=self.name,
            _namespace=namespace,
            _target=self.target,
            args=self.args,
            kwargs=self.kwargs,
            target_fn=self.target_fn,
        )

    def resolves(self, config: Dict[str, Any]) -> bool:
        """Returns whether the Applicable resolves with the given config

        :param config: Configuration to check
        :return: Whether the Applicable resolves with the given config
        """
        for resolver in self.resolvers:
            if not resolver(config):
                return False
        return True

    def named(self, name: str, namespace: NamespaceType = ...) -> "Applicable":
        """Names the function application. This has the following rules:
        1. The name will be the name passed in, this is required
        2. If the namespace is `None`, then there will be no namespace
        3. If the namespace is `...`, then the namespace will be the namespace that already exists, usually the name of the
        function that this is decorating. This is an odd case -- but it helps if you have
        multiple of the same type of operations that you want to apply across different nodes,
        or in the case of a parameterization (which is not yet supported).

        :param name: Name of the node to be created
        :param namespace: Namespace of the node to be created -- currently only single-level namespaces are supported
        :return: The applicable with the new name
        """
        return Applicable(
            fn=self.fn,
            _resolvers=self.resolvers,
            _name=name if name is not None else self.name,
            _namespace=(
                None
                if namespace is None
                else (namespace if namespace is not ... else self.namespace)
            ),
            _target=self.target,
            args=self.args,
            kwargs=self.kwargs,
            target_fn=self.target_fn,
        )

    # on_input / on_output are the same but here for naming convention
    # I know there is a way to dynamically resolve this to revert to a common function
    # just can't remember it now or find it online...
    # TODO: adding the option to select target parameter for each transform
    # def on_input(self, target: base.TargetType) -> "Applicable":
    #     """Add Target on a single function level.

    #     This determines to which node(s) it will applies. Should match the same naming convention
    #     as the NodeTransorfmLifecycle child class (for example NodeTransformer).

    #     :param target: Which node(s) to apply on top of
    #     :return: The Applicable with specified target
    #     """
    #     return Applicable(
    #         fn=self.fn,
    #         _resolvers=self.resolvers,
    #         _name=self.name,
    #         _namespace=self.namespace,
    #         _target=target if target is not None else self.target,
    #         args=self.args,
    #         kwargs=self.kwargs,
    #         target_fn=self.target_fn,
    #     )

    def on_output(self, target: base.TargetType) -> "Applicable":
        """Add Target on a single function level.

        This determines to which node(s) it will applies. Should match the same naming convention
        as the NodeTransorfmLifecycle child class (for example NodeTransformer).

        :param target: Which node(s) to apply on top of
        :return: The Applicable with specified target
        """
        return Applicable(
            fn=self.fn,
            _resolvers=self.resolvers,
            _name=self.name,
            _namespace=self.namespace,
            _target=target if target is not None else self.target,
            args=self.args,
            kwargs=self.kwargs,
            target_fn=self.target_fn,
        )

    def get_config_elements(self) -> List[str]:
        """Returns the config elements that this Applicable uses"""
        out = []
        for resolver in self.resolvers:
            out.extend(resolver.optional_config)
        return out

    def validate(self, chain_first_param: bool, allow_custom_namespace: bool):
        """Validates that the Applicable function can be applied given the
        set of args/kwargs passed in. This says that:

        1. The signature binds appropriately
        2. If we chain the first parameter, it is not present in the function

        Note that this is currently restrictive. We only support hamilton-friendly functions. Furthermore,
        this logic is slightly duplicated from `@does` above. We will be suporting more function shapes (in
        both this and `@does`) over time, and also combining the logic between the two for
        validating/binding signatures.

        :param chain_first_param: Whether we chain the first parameter
        :raises InvalidDecoratorException if the function cannot be applied
        :return:
        """
        args = ((...,) if chain_first_param else ()) + tuple(self.args)  # dummy argument at first
        sig = inspect.signature(self.fn)
        if len(sig.parameters) == 0:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} has no parameters. "
                f"You cannot apply a function with no parameters."
            )
        invalid_args = [
            item
            for item in inspect.signature(self.fn).parameters.values()
            if item.kind
            not in {
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY,
            }
        ]
        if len(invalid_args) > 0:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} has invalid parameters. "
                "You cannot apply a function with parameters that are not keyword-friendly. "
                f"The following parameters are not keyword-friendly: {invalid_args}"
            )
        try:
            sig.bind(*args, **self.kwargs)
        except TypeError as e:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} cannot be applied with the following args: {self.args} "
                f"and the following kwargs: {self.kwargs}"
            ) from e
        if len(sig.parameters) == 0:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} has no parameters. "
                "You cannot apply a function with no parameters."
            )
        if self.namespace is not ... and not allow_custom_namespace:
            raise base.InvalidDecoratorException(
                "Currently, setting namespace globally inside "
                "pipe(...)/flow(...) is not compatible with setting namespace "
                "for a step(...) call."
            )
        try:
            node.Node.from_fn(self.fn)
        except ValueError as e:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} cannot be applied with the following args: {self.args} "
                f"and the following kwargs: {self.kwargs}. See documentation on pipe(), the function "
                "shapes are currently restrictive to anything with named kwargs (either kwarg-only or positional/kwarg arguments), "
                "and must be typed. If you need functions that don't have these requirements, please reach out to the Hamilton team."
                "Current workarounds are to define a wrapper function that assigns types with the proper keyword-friendly arguments."
            ) from e

    def resolve_namespace(self, default_namespace: str) -> Tuple[str, ...]:
        """Resolves the namespace -- see rules in `named` for more details.

        :param default_namespace: namespace to use as a default if we do not wish to override it
        :return: The namespace to use, as a tuple (hierarchical)
        """
        return (
            (default_namespace,)
            if self.namespace is ...
            else (self.namespace,)
            if self.namespace is not None
            else ()
        )

    def bind_function_args(
        self, current_param: Optional[str]
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Binds function arguments, given current, chained parameeter

        :param current_param: Current, chained parameter. None, if we're not chaining.
        :return: A tuple of (upstream_inputs, literal_inputs)
        """
        args_to_bind = self.args
        if current_param is not None:
            args_to_bind = (source(current_param),) + args_to_bind
        kwargs_to_bind = self.kwargs
        fn_signature = inspect.signature(self.fn)
        bound_signature = fn_signature.bind(*args_to_bind, **kwargs_to_bind)
        all_kwargs = {**bound_signature.arguments, **bound_signature.kwargs}
        upstream_inputs = {}
        literal_inputs = {}
        # TODO -- restrict to ensure that this covers *all* dependencies
        # TODO -- bind to parameters using args
        for dep, value in all_kwargs.items():
            if isinstance(value, UpstreamDependency):
                upstream_inputs[dep] = value.source
            elif isinstance(value, LiteralDependency):
                literal_inputs[dep] = value.value
            else:
                literal_inputs[dep] = value

        return upstream_inputs, literal_inputs


def step(
    fn, *args: Union[SingleDependency, Any], **kwargs: Union[SingleDependency, Any]
) -> Applicable:
    """Applies a function to for a node (or a subcomponent of a node).
    See documentation for `pipe` to see how this is used.

    :param fn: Function to use. Must be validly called as f(**kwargs), and have a 1:1 mapping of kwargs to parameters.
    :param args: Args to pass to the function -- although these cannot be variable/position-only arguments, they can be
    positional arguments. If these are not source/value, they will be converted to a value (a literal)
    :param kwargs: Kwargs to pass to the function. These can be source/value, or they can be literals. If they are literals,
    they will be converted to a value (a literal)
    :return: an applicable with the function applied
    """
    return Applicable(fn=fn, _resolvers=[], args=args, kwargs=kwargs)


# TODO: In case of multiple parameter targets we want to safeguard that it is a clear target distribution
# class MissingTargetError(Exception):
#     """When setting target make sure it is clear which transform targets which node.

#     This is a safeguard, because the default behavior may not apply if targets are partially set
#     and we do not want to make assumptions what the user meant.
#     """

#     pass


class pipe_input(base.NodeInjector):
    """Running a series of transformations on the input of the function.

    To demonstrate the rules for chaining nodes, we'll be using the following example. This is
    using primitives to demonstrate, but as hamilton is just functions of any python objects, this works perfectly with
    dataframes, series, etc...


    .. code-block:: python
        :name: Simple @pipe_input example

        from hamilton.function_modifiers import step, pipe_input, value, source


        def _add_one(x: int) -> int:
            return x + 1


        def _sum(x: int, y: int) -> int:
            return x + y


        def _multiply(x: int, y: int, z: int = 10) -> int:
            return x * y * z


        @pipe_input(
            step(_add_one),
            step(_multiply, y=2),
            step(_sum, y=value(3)),
            step(_multiply, y=source("upstream_node_to_multiply")),
        )
        def final_result(upstream_int: int) -> int:
            return upstream_int

    .. code-block:: python
        :name: Equivalent example with no @pipe_input, nested

        upstream_int = ...  # result from upstream
        upstream_node_to_multiply = ...  # result from upstream

        output = final_result(
            _multiply(
                _sum(
                    _multiply(
                        _add_one(upstream_int),
                        y=2
                    ),
                    y=3
                ),
                y=upstream_node_to_multiply
            )
        )

    .. code-block:: python
        :name: Equivalent example with no @pipe_input, procedural

        upstream_int = ...  # result from upstream
        upstream_node_to_multiply = ...  # result from upstream

        one_added = _add_one(upstream_int)
        multiplied = _multiply(one_added, y=2)
        summed = _sum(multiplied, y=3)
        multiplied_again = _multiply(summed, y=upstream_node_to_multiply)
        output = final_result(multiplied_again)


    Note that functions must have no position-only arguments (this is rare in python, but hamilton does not handle these).
    This basically means that the functions must be defined similarly to ``def fn(x, y, z=10)`` and not ``def fn(x, y, /, z=10)``.
    In fact, all arguments must be named and "kwarg-friendly", meaning that the function can happily be called with ``**kwargs``,
    where kwargs are some set of resolved upstream values. So, no ``*args`` are allowed, and ``**kwargs`` (variable keyword-only) are not
    permitted. Note that this is not a design limitation, rather an implementation detail -- if you feel like you need this, please
    reach out.

    Furthermore, the function should be typed, as a Hamilton function would be.

    One has three ways to tune the shape/implementation of the subsequent nodes:

    1. ``when``/``when_not``/``when_in``/``when_not_in`` -- these are used to filter the application of the function.
        This is valuable to reflect if/else conditions in the structure of the DAG, pulling it out of functions, rather
        than buried within the logic itself. It is functionally equivalent to ``@config.when``.

        For instance, if you want to include a function in the chain only when a config parameter is set to a certain value, you can do:

        .. code-block:: python

            @pipe_input(
                step(_add_one).when(foo="bar"),
                step(_add_two, y=source("other_node_to_add").when(foo="baz"),
            )
            def final_result(upstream_int: int) -> int:
                return upstream_int

        This will only apply the first function when the config parameter ``foo`` is set to ``bar``, and the second when it is set to ``baz``.

    2. ``named`` -- this is used to name the node. This is useful if you want to refer to intermediate results.
        If this is left out, hamilton will automatically name the functions in a globally unique manner. The names of
        these functions will not necessarily be stable/guaranteed by the API, so if you want to refer to them, you should use ``named``.
        The default namespace will always be the name of the decorated function (which will be the last node in the chain).

        ``named`` takes in two parameters -- required is the ``name`` -- this will assign the nodes with a single name and *no* global namespace.
        For instance:

        .. code-block:: python

            @pipe_input(
                step(_add_one).named("a"),
                step(_add_two, y=source("upstream_node")).named("b"),
            )
            def final_result(upstream_int: int) -> int:
                return upstream_int

        The above will create two nodes, ``a`` and ``b``. ``a`` will be the result of ``_add_one``, and ``b`` will be the result of ``_add_two``.
        ``final_result`` will then be called with the output of ``b``. Note that, if these are part of a namespaced operation (a subdag, in particular),
        they *will* get the same namespace as the subdag.

        The second parameter is ``namespace``. This is used to specify a namespace for the node. This is useful if you want
        to either (a) ensure that the nodes are namespaced but share a common one to avoid name clashes (usual case), or (b)
        if you want a custom namespace (unusual case). To indicate a custom namespace, one need simply pass in a string.

        To indicate that a node should share a namespace with the rest of the step(...) operations in a pipe, one can pass in ``...`` (the ellipsis).

        .. code-block:: python
          :name: Namespaced step


            @pipe_input(
                step(_add_one).named("a", namespace="foo"),  # foo.a
                step(_add_two, y=source("upstream_node")).named("b", namespace=...),  # final_result.b
            )
            def final_result(upstream_int: int) -> int:
                return upstream_int

        Note that if you pass a namespace argument to the ``pipe_input`` function, it will set the namespace on each step operation.
        This is useful if you want to ensure that all the nodes in a pipe have a common namespace, but you want to rename them.

        .. code-block:: python
            :name: pipe_input with globally applied namespace

            @pipe_input(
                step(_add_one).named("a"), # a
                step(_add_two, y=source("upstream_node")).named("b"), # foo.b
                namespace=..., # default -- final_result.a and final_result.b, OR
                namespace=None, # no namespace -- a and b are exposed as that, OR
                namespace="foo", # foo.a and foo.b
            )
            def final_result(upstream_int: int) -> int:
                return upstream_int

        In all likelihood, you should not be using this, and this is only here in case you want to expose a node for
        consumption/output later. Setting the namespace in individual nodes as well as in ``pipe_input`` is not yet supported.

    3. ``on_input`` -- this selects which input we will run the pipeline on.
        In case ``on_input`` is set to None (default), we apply ``pipe_input`` on the first parameter. Let us know if you wish to expand to other use-cases.
        You can track the progress on this topic via: https://github.com/DAGWorks-Inc/hamilton/issues/1177

        The following would apply function *_add_one* and *_add_two* to ``p2``:

        .. code-block:: python

            @pipe_input(
                step(_add_one)
                step(_add_two, y=source("upstream_node")),
                on_input = "p2"
            )
            def final_result(p1: int, p2: int, p3: int) -> int:
                return upstream_int

        .. |
            THIS IS COMMENTED OUT, I.E. SPHINX WILL NOT AUTODOC IT, HERE IN CASE WE ENABLE MULTIPLE PARAMETER TARGETS
            For extra control in case of multiple function arguments (parameters), we can also specify the target parameter that we wish to transform.
            In case ``on_input`` is set to None (default), we apply ``pipe_input`` on the first parameter only. If ``on_input`` is set for a specific transform
            make sure the other ones are also set either through a global setting or individually, otherwise it is unclear which transforms target which parameters.

            The following applies *_add_one* to ``p1``, ``p3`` and *_add_two* to ``p2``

            .. code-block:: python

                @pipe_input(
                    step(_add_one).on_input(["p1","p3"])
                    step(_add_two, y=source("upstream_node")).on_input("p2")
                )
                def final_result(p1: int, p2: int, p3: int) -> int:
                    return upstream_int

            We can also do this on the global level to set for all transforms a target parameter.

            Lastly, a mixture of global and local is possible, where the global selects the target parameters for
            all transforms and we can select individual transforms to also target more parameters.
            The following would apply function *_add_one* to all ``p1``, ``p2``, ``p3`` and *_add_two* also on ``p2``

            .. code-block:: python

                @pipe_input(
                    step(_add_one).on_input(["p1","p3"])
                    step(_add_two, y=source("upstream_node")),
                    on_input = "p2"
                )
                def final_result(p1: int, p2: int, p3: int) -> int:
                    return upstream_int
            | replace:: \
    """

    def __init__(
        self,
        *transforms: Applicable,
        namespace: NamespaceType = ...,
        on_input: base.TargetType = None,
        collapse=False,
        _chain=False,
    ):
        """Instantiates a ``@pipe_input`` decorator.

        :param transforms: step transformations to be applied, in order
        :param namespace: namespace to apply to all nodes in the pipe. This can be "..." (the default), which resolves to the name of the decorated function, None (which means no namespace), or a string (which means that all nodes will be namespaced with that string). Note that you can either use this *or* namespaces inside ``pipe_input()``...
        :param on_input: setting the target parameter for all steps in the pipe. Leave empty to select only the first argument.
        :param collapse: Whether to collapse this into a single node. This is not currently supported.
        :param _chain: Whether to chain the first parameter. This is the only mode that is supported. Furthermore, this is not externally exposed. ``@flow`` will make use of this.
        """
        if on_input is not None:
            if not isinstance(on_input, str):
                raise NotImplementedError(
                    "on_input currently only supports a single target parameter specified by a string. "
                    "Please reach out if you want a more flexible option in the feature."
                )
        base.NodeTransformer._early_validate_target(target=on_input, allow_multiple=True)

        self.transforms = transforms
        self.collapse = collapse
        self.chain = _chain
        self.namespace = namespace
        self.target = [on_input]

        # TODO: for multiple target parameter case
        # if isinstance(on_input, str):  # have to do extra since strings are collections in python
        #     self.target = [on_input]
        # elif isinstance(on_input, Collection):
        #     self.target = on_input
        # else:
        #     self.target = [on_input]

        if self.collapse:
            raise NotImplementedError(
                "Collapsing step() functions as one node is not yet implemented for pipe_input(). Please reach out if you want this feature."
            )

        if self.chain:
            raise NotImplementedError("@flow() is not yet supported -- this is ")

    def _distribute_transforms_to_parameters(
        self, params: Dict[str, Type[Type]]
    ) -> Dict[str, List[Applicable]]:
        """Resolves target option on the transform level.
        Adds option that we can decide for each applicable which input parameter it will target on
        top of the global target (if it is set).

        We create a hash for each target parameter with a list of transforms that will be applied
        to this parameter.

        :params: Available input parameters of the function
        :return: A dictionary mapping between selected parameters and list of transforms
        """

        selected_transforms = defaultdict(list)
        for param in params:
            if param in self.target:
                selected_transforms[param].extend(self.transforms)
            # TODO: in case of multiple parameters we can set individual targets and resolve them here
            # for transform in self.transforms:
            #     target = transform.target
            #     # In case there is no target set on applicable we assign global target
            #     if target is None:
            #         target = self.target
            #     elif isinstance(target, str):  # user selects single target via string
            #         target = [target]
            #         target.extend(self.target)
            #     elif isinstance(target, Collection):  # user inputs a list of targets
            #         target.extend(self.target)

            # if param in target:
            #     selected_transforms[param].append(transform)

        return selected_transforms

    def _create_valid_parameters_transforms_mapping(
        self, mapping: Dict[str, List[Applicable]], fn: Callable, params: Dict[str, Type[Type]]
    ) -> Dict[str, List[Applicable]]:
        """Checks for a valid distribution of transforms to parameters."""
        sig = inspect.signature(fn)
        param_names = []
        for param in list(sig.parameters.values()):
            param_names.append(param.name)
            # use the name of the parameter to determine the first node
            # Then wire them all through in order
            # if it resolves, great
            # if not, skip that, pointing to the previous
            # Create a node along the way

        if not mapping:
            # This reverts back to legacy chaining through first parameter and checks first parameter
            first_parameter = param_names[0]
            if first_parameter not in params:
                raise base.InvalidDecoratorException(
                    f"Function: {fn.__name__} has a first parameter that is not a dependency. "
                    f"@pipe requires the parameter names to match the function parameters. "
                    f"Thus it might not be compatible with some other decorators"
                )
            mapping[first_parameter] = self.transforms
        # TODO: validate that all transforms have a target in case multiple parameters targeted
        # else:
        #     # in case we set target this checks that each transform has at least one target parameter
        #     transform_set = []
        #     for param in mapping:
        #         transform_set.extend(mapping[param])
        #     transform_set = set(transform_set)
        #     if len(transform_set) != len(self.transforms):
        #         raise MissingTargetError(
        #             "The on_input settings are unclear. Please make sure all transforms "
        #             "either have specified individually or globally a target or there is "
        #             "no on_input usage."
        #         )

        # similar to above we check that the target parameter is among the actual function parameters
        if next(iter(mapping)) not in param_names:
            raise base.InvalidDecoratorException(
                f"Function: {fn.__name__} with parameters {param_names} does not a have "
                f"dependency {next(iter(mapping))}. @pipe_input requires the parameter "
                f"names to match the function parameters. Thus it might not be compatible "
                f"with some other decorators."
            )

        return mapping

    def _resolve_namespace(
        self,
        param: str,
    ) -> str:
        """Add parameter name to namespace.
        In case we pipe_input on multiple parameters we have to duplicate nodes to be able to chain
        them together for each argument and they have to have different names.
        """
        if self.namespace is ... or self.namespace is None:
            return param
        else:
            return f"{self.namespace}_{param}"

    def inject_nodes(
        self, params: Dict[str, Type[Type]], config: Dict[str, Any], fn: Callable
    ) -> Tuple[List[node.Node], Dict[str, str]]:
        """Injects nodes into the graph. This creates a node for each pipe() step,
        then reassigns the inputs to pass it in."""

        parameters_transforms_mapping = self._distribute_transforms_to_parameters(params=params)
        parameters_transforms_mapping = self._create_valid_parameters_transforms_mapping(
            mapping=parameters_transforms_mapping, fn=fn, params=params
        )

        total_nodes = []
        total_rename_maps = {}

        for param in parameters_transforms_mapping:
            # If only single parameter we revert to previous namespace convention since no duplication issues
            # This also ensures backwards compatibility
            if len(parameters_transforms_mapping) == 1:
                namespace = self.namespace
            else:
                namespace = self._resolve_namespace(param=param)

            # Chaining gets done by linking the specified argument of each node
            nodes, current_param = chain_transforms(
                target_arg=param,
                transforms=parameters_transforms_mapping[param],
                namespace=namespace,
                config=config,
                fn=fn,
            )
            total_nodes.extend(nodes)
            total_rename_maps.update({param: current_param})

        return total_nodes, total_rename_maps  # rename to ensure it all works

    def validate(self, fn: Callable):
        """Validates the the individual steps work together."""
        for applicable in self.transforms:
            applicable.validate(
                chain_first_param=True, allow_custom_namespace=self.namespace is ...
            )
        # TODO -- validate that the types match on the chain (this is de-facto done later)

    def optional_config(self) -> Dict[str, Any]:
        """Declares the optional configuration keys for this decorator.
        These are configuration keys that can be used by the decorator, but are not required.
        Along with these we have *defaults*, which we will use to pass to the config.

        :return: The optional configuration keys with defaults. Note that this will return None
        if we have no idea what they are, which bypasses the configuration filtering we use entirely.
        This is mainly for the legacy API.
        """
        out = {}
        for applicable in self.transforms:
            for resolver in applicable.resolvers:
                out.update(resolver.optional_config)
        return out


@deprecated(
    warn_starting=(1, 20, 0),
    fail_starting=(2, 0, 0),
    use_this=pipe_input,
    explanation="pipe has been replaced with pipe_input -- a clearer name since "
    "we also added pipe_output with complimentary functionality.",
    current_version=(1, 77, 0),
    migration_guide="https://hamilton.dagworks.io/en/latest/reference/decorators/",
)
class pipe(pipe_input):
    def __init__(
        self,
        *transforms: Applicable,
        namespace: NamespaceType = ...,
        on_input: base.TargetType = None,
        collapse=False,
        _chain=False,
    ):
        super(pipe, self).__init__(
            *transforms,
            namespace=namespace,
            on_input=on_input,
            collapse=False,
            _chain=False,
        )


# # TODO -- implement flow!
# class flow(pipe):
#     """flow() is a more flexible, power-user version of `pipe`. The rules are largely similar, with a few key differences:
#
#     1. The first parameter is not passed through -- the user is responsible for passing all parameters into a function
#     2. The final function can depend on any of the prior functions -- it will declare those as inputs using the input parameters. These will
#     be seen as inputs, regardless of namespace.
#
#     This means that `flow` can be used to construct *any* DAG -- this is why its a power-user capability. Before we dig into some examples,
#     a quick note on when to use it:
#
#     flow is meant to procedurally specify a subcomponent of the DAG. While Hamilton encourage declarative (not procedural) DAGs, there are
#     certain cases where you may find yourself wanting to more dynamically construct a DAG, for certain subcomponents. This is where
#     `flow` comes in. As we will show later in this doc, this can be very powerful when combined with `resolve` to build configuration-driven
#     DAGs. Again, however, this is meant to be a subset of a declarative DAG -- procedurally defined subdags can help with flexibility, but
#     should not be overused.
#
#     Now, let's get to some examples:
#
#     TODO -- basic example
#
#     TODO -- example with resolve
#
#     TODO -- examples with namespacing
#     """
#
#     def __init__(self, *transforms: Applicable, collapse=False):
#         super(flow, self).__init__(*transforms, collapse=collapse, _chain=False)


class SingleTargetError(Exception):
    """We prohibit the target to be raise both globally and locally.

    Decorators that transform the output of a node can be set to transform only
    a certain output node (useful with extract_columns / extract_fields). Some decorators
    can group multiple transforms and we can set that certain output node either for all of them
    or for each individually.

    This is a safeguard, because when you set the global target it creates a subset of those nodes and
    if the local target is outside of that subset it gets ignore (opposed to the logical assumption that
    it can override the global one). So we disable that case.
    """

    pass


class pipe_output(base.NodeTransformer):
    """Running a series of transformation on the output of the function.

    The decorated function declares the dependency, the body of the function gets executed, and then
    we run a series of transformations on the result of the function specified by ``pipe_output``.

    If we have nodes **A --> B --> C** in the DAG and decorate **B** with ``pipe_output`` like

    .. code-block:: python
        :name: Simple @pipe_output example

        @pipe_output(
            step(B1),
            step(B2)
        )
        def B(...):
            return ...

    we obtain the new DAG **A --> B.raw --> B1 --> B2 --> B --> C**, where we can think of the **B.raw --> B1 --> B2 --> B** as a "pipe" that takes the raw output of **B**, applies to it
    **B1**, takes the output of **B1** applies to it **B2** and then gets renamed to **B** to re-connect to the rest of the DAG.

    The rules for chaining nodes are the same as for ``pipe_input``.

    For extra control in case of multiple output nodes, for example after ``extract_field``/ ``extract_columns`` we can also specify the output node that we wish to mutate.
    The following apply *A* to all fields while *B* only to ``field_1``

    .. code-block:: python
        :name: Simple @pipe_output example targeting specific nodes

        @extract_columns("col_1", "col_2")
        def A(...):
            return ...

        def B(...):
            return ...


         @pipe_output(
            step(A),
            step(B).on_output("field_1"),
        )
        @extract_fields(
                {"field_1":int, "field_2":int, "field_3":int}
        )
        def foo(a:int)->Dict[str,int]:
            return {"field_1":1, "field_2":2, "field_3":3}

    We can also do this on the global level (but cannot do on both levels at the same time). The following would apply function *A* and function *B* to only ``field_1`` and ``field_2``

    .. code-block:: python
        :name: Simple @pipe_output targeting specific nodes local

        @pipe_output(
            step(A),
            step(B),
            on_output = ["field_1","field_2]
        )
        @extract_fields(
                {"field_1":int, "field_2":int, "field_3":int}
        )
        def foo(a:int)->Dict[str,int]:
            return {"field_1":1, "field_2":2, "field_3":3}
    """

    @classmethod
    def _validate_single_target_level(cls, target: base.TargetType, transforms: Tuple[Applicable]):
        """We want to make sure that target gets applied on a single level.
        Either choose for each step individually what it targets or set it on the global level where
        all steps will target the same node(s).
        """
        if target is not None:
            for transform in transforms:
                if transform.target is not None:
                    raise SingleTargetError("Cannot have target set on pipe_output and step level.")

    def __init__(
        self,
        *transforms: Applicable,
        namespace: NamespaceType = ...,
        on_output: base.TargetType = None,
        collapse=False,
        _chain=False,
    ):
        """Instantiates a ``@pipe_output`` decorator.

        Warning: if there is a global pipe_output target, the individual ``step(...).target`` would only choose
        from the subset pre-selected from the global pipe_output target. We have disabled this for now to avoid
        confusion. Leave global pipe_output target empty if you want to choose between all the nodes on the individual step level.

        :param transforms: step transformations to be applied, in order
        :param namespace: namespace to apply to all nodes in the pipe. This can be "..." (the default), which resolves to the name of the decorated function, None (which means no namespace), or a string (which means that all nodes will be namespaced with that string). Note that you can either use this *or* namespaces inside ``pipe_output()``...
        :param on_output: setting the target node for all steps in the pipe. Leave empty to select all the output nodes.
        :param collapse: Whether to collapse this into a single node. This is not currently supported.
        :param _chain: Whether to chain the first parameter. This is the only mode that is supported. Furthermore, this is not externally exposed. ``@flow`` will make use of this.
        """
        pipe_output._validate_single_target_level(target=on_output, transforms=transforms)

        if on_output == ...:
            raise ValueError(
                "Cannot apply Elipsis(...) to on_output. Use None, single string or list of strings."
            )

        super(pipe_output, self).__init__(target=on_output)
        self.transforms = transforms
        self.collapse = collapse
        self.chain = _chain
        self.namespace = namespace

        if self.collapse:
            raise NotImplementedError(
                "Collapsing step() functions as one node is not yet implemented for pipe(). Please reach out if you want this feature."
            )

        if self.chain:
            raise NotImplementedError("@flow() is not yet supported -- this is ")

    def _filter_individual_target(self, node_):
        """Resolves target option on the transform level.
        Adds option that we can decide for each applicable which output node it will target.

        :param node_: The current output node.
        :return: The set of transforms that target this node
        """
        selected_transforms = []
        for transform in self.transforms:
            target = transform.target
            if isinstance(target, str):  # user selects single target via string
                if node_.name == target:
                    selected_transforms.append(transform)
            elif isinstance(target, Collection):  # user inputs a list of targets
                if node_.name in target:
                    selected_transforms.append(transform)
            else:  # for target=None (default) we include all sink nodes
                selected_transforms.append(transform)

        return tuple(selected_transforms)

    def transform_node(
        self, node_: node.Node, config: Dict[str, Any], fn: Callable
    ) -> Collection[node.Node]:
        """Injects nodes into the graph.

        We create a copy of the original function and rename it to `function_name.raw` to be the
        initial node. Then we create a node for each step in `post-pipe` and chain them together.
        The last node is an identity to the previous one with the original name `function_name` to
        represent an exit point of `pipe_output`.
        """
        transforms = self._filter_individual_target(node_)
        if len(transforms) < 1:
            # in case no functions in pipeline we short-circuit and return the original node
            return [node_]

        if self.namespace is None:
            _namespace = None
        elif self.namespace is ...:
            _namespace = node_.name
        else:
            _namespace = self.namespace

        # We pick a reserved prefix that ovoids clashes with user defined functions / nodes
        original_node = node_.copy_with(name=f"{node_.name}.raw")

        def __identity(foo: Any) -> Any:
            return foo

        transforms = transforms + (step(__identity).named(fn.__name__),)
        nodes, _ = chain_transforms(
            target_arg=original_node.name,
            transforms=transforms,
            namespace=_namespace,  # self.namespace,
            config=config,
            fn=fn,
        )

        last_node = nodes[-1].copy_with(name=f"{node_.name}", typ=nodes[-2].type)

        out = [original_node]
        out.extend(nodes[:-1])
        out.append(last_node)
        return out

    def validate(self, fn: Callable):
        """Validates the the individual steps work together."""
        for applicable in self.transforms:
            applicable.validate(
                chain_first_param=True, allow_custom_namespace=self.namespace is ...
            )
        # TODO -- validate that the types match on the chain (this is de-facto done later)

    def optional_config(self) -> Dict[str, Any]:
        """Declares the optional configuration keys for this decorator.
        These are configuration keys that can be used by the decorator, but are not required.
        Along with these we have *defaults*, which we will use to pass to the config.

        :return: The optional configuration keys with defaults. Note that this will return None
        if we have no idea what they are, which bypasses the configuration filtering we use entirely.
        This is mainly for the legacy API.
        """
        out = {}
        for applicable in self.transforms:
            for resolver in applicable.resolvers:
                out.update(resolver.optional_config)
        return out


def chain_transforms(
    target_arg: str,
    transforms: List[Applicable],
    namespace: str,
    config: Dict[str, Any],
    fn: Callable,
):
    """Chaining nodes together sequentially through the a specified argument.

    :param target_arg: assigning the name of the specified argument of the first node in chain
    :param transforms: step transformations to be applied, in order
    :param namespace: namespace to apply to all nodes. This can be "..." (the default), which resolves to the name of the decorated function, None (which means no namespace), or a string (which means that all nodes will be namespaced with that string)
    :param config: Configuration to use -- this can be specified in the decorator
    :param fn: initial function that was decorated

    :return: A list of nodes that have been chained together through the specified argument.
    """

    fn_count = Counter()
    nodes = []
    for applicable in transforms:
        if namespace is not ...:
            applicable = applicable.namespaced(
                namespace=namespace
            )  # we reassign the global namespace
        if applicable.resolves(config):
            fn_name = applicable.fn.__name__
            postfix = "" if fn_count[fn_name] == 0 else f"_{fn_count[fn_name]}"
            node_name = (
                applicable.name
                if applicable.name is not None
                else f"with{('_' if not fn_name.startswith('_') else '') + fn_name}{postfix}"
            )
            raw_node = node.Node.from_fn(
                applicable.fn,
                f"with{('_' if not fn_name.startswith('_') else '') + fn_name}{postfix}",
            )
            node_namespace = applicable.resolve_namespace(fn.__name__)
            raw_node = raw_node.copy_with(namespace=node_namespace, name=node_name)
            # TODO -- validate that the first parameter is the right type/all the same
            fn_count[fn_name] += 1
            upstream_inputs, literal_inputs = applicable.bind_function_args(target_arg)
            nodes.append(
                raw_node.reassign_inputs(
                    input_names=upstream_inputs,
                    input_values=literal_inputs,
                )
            )
            target_arg = raw_node.name
    return nodes, target_arg


def apply_to(fn_: Union[Callable, str], **mutating_fn_kwargs: Union[SingleDependency, Any]):
    """Creates an applicable placeholder with potential kwargs that will be applied to a node (or a subcomponent of a node).
    See documentation for ``mutate`` to see how this is used. It de facto allows a postponed ``step``.

    We pass fn=None here as this will be the function we are decorating and need to delay passing it in. The target
    function is the one we wish to mutate and we store it for later access.

    :param fn: Function the applicable will be applied to
    :param mutating_fn_kwargs: Kwargs (**kwargs) to pass to the mutator function. Must be validly called as f(**kwargs), and have a 1:1 mapping of kwargs to parameters.
    :return: an applicable placeholder with the target function
    """
    return Applicable(fn=None, args=(), kwargs=mutating_fn_kwargs, target_fn=fn_, _resolvers=[])


class NotSameModuleError(Exception):
    """Limit the use of a decorator on functions from the same module.

    Some decorators have the ability to transform also other functions than the one they are decorating (for example mutate).
    This ensures that all the functions are located within the same module.
    """

    def __init__(self, fn: Callable, target_fn: Callable):
        super().__init__(
            f"The functions have to be in the same module... "
            f"The target function {target_fn.__name__} is in module {target_fn.__module__} and "
            f"the mutator function {fn.__name__} is in module {fn.__module__}./n"
            "Use power user setting to disable this restriction."
        )


class mutate:
    """Running a transformation on the outputs of a series of functions.

    This is closely related to ``pipe_output`` as it effectively allows you to run transformations on the output of a node without touching that node.
    We choose which target functions we wish to mutate by the transformation we are decorating. For now, the target functions, that will be mutated,
    have to be in the same module (come speak to us if you need this capability over multiple modules).

    We suggest you define them with an prefixed underscore to only have them displayed in the `transform pipeline` of the target node.

    If we wish to apply ``_transform1`` to the output of **A** and **B** and ``_transform2`` only to the output
    of node **B**, we can do this like

    .. code-block:: python
        :name: Simple @mutate example

        def A(...):
            return ...

        def B(...):
            return ...

        @mutate(A, B)
        def _transform1(...):
            return ...

        @mutate(B)
        def _transform2(...):
            return ...

    we obtain the new pipe-like subDAGs **A.raw --> _transform1 --> A** and **B.raw --> _transform1 --> _transform2 --> B**,
    where the behavior is the same as ``pipe_output``.

    While it is generally reasonable to use ``pipe_output``, you should consider ``mutate`` in the following scenarios:

    1. Loading data and applying pre-cleaning step.
    2. Feature engineering via joining, filtering, sorting, etc.
    3. Experimenting with different transformations across nodes by selectively turning transformations on / off.

    We assume the first argument of the decorated function to be the output of the function we are targeting.
    For transformations with multiple arguments you can use key word arguments coupled with ``step`` or ``value``
    the same as with other ``pipe``-family decorators

    .. code-block:: python
        :name: Simple @mutate example with multiple arguments

        @mutate(A, B, arg2=step('upstream_node'), arg3=value(some_literal), ...)
        def _transform1(output_from_target:correct_type, arg2:arg2_type, arg3:arg3_type,...):
            return ...

    You can also select individual args that will be applied to each target node by adding ``apply_to(...)``

    .. code-block:: python
        :name: Simple @mutate example with multiple arguments allowing individual actions

        @mutate(
                apply_to(A, arg2=step('upstream_node_1'), arg3=value(some_literal_1)),
                apply_to(B, arg2=step('upstream_node_2'), arg3=value(some_literal_2)),
                )
        def _transform1(output_from_target:correct_type, arg2:arg2_type, arg3:arg3_type, ...):
            return ...

    In case of multiple output nodes, for example after ``extract_field`` / ``extract_columns`` we can also specify the output node that we wish to mutate.
    The following would mutate all columns of *A* individually while in the case of function *B* only ``field_1``

    .. code-block:: python
        :name: @mutate example targeting specific nodes local

        @extract_columns("col_1", "col_2")
        def A(...):
            return ...

        @extract_fields(
                {"field_1":int, "field_2":int, "field_3":int}
        )
        def B(...):
            return ...

        @mutate(
            apply_to(A),
            apply_to(B).on_output("field_1"),
            )

        def foo(a:int)->Dict[str,int]:
            return {"field_1":1, "field_2":2, "field_3":3}
    """

    def __init__(
        self,
        *target_functions: Union[Applicable, Callable],
        collapse: bool = False,
        _chain: bool = False,
        **mutating_function_kwargs: Union[SingleDependency, Any],
    ):
        """Instantiates a ``mutate`` decorator.

        We assume the first argument of the decorated function to be the output of the function we are targeting.

        :param target_functions: functions we wish to mutate the output of
        :param collapse: Whether to collapse this into a single node. This is not currently supported.
        :param _chain: Whether to chain the first parameter. This is the only mode that is supported. Furthermore, this is not externally exposed. ``@flow`` will make use of this.
        :param `**mutating_function_kwargs`: other kwargs that the decorated function has. Must be validly called as ``f(**kwargs)``, and have a 1-to-1 mapping of kwargs to parameters. This will be applied for all ``target_functions``, unless ``apply_to`` already has the mutator function kwargs, in which case it takes those.
        """
        self.collapse = collapse
        self.chain = _chain
        # keeping it here once it gets implemented maybe nice to have options
        if self.collapse:
            raise NotImplementedError(
                "Collapsing functions as one node is not yet implemented for mutate(). Please reach out if you want this feature."
            )
        if self.chain:
            raise NotImplementedError("@flow() is not yet supported -- this is ")

        self.remote_applicables = tuple(
            [apply_to(fn) if isinstance(fn, Callable) else fn for fn in target_functions]
        )
        self.mutating_function_kwargs = mutating_function_kwargs

        # Cross module will require some thought so we are restricting mutate to single module for now
        self.restrict_to_single_module = True

    def validate_same_module(self, mutating_fn: Callable):
        """Validates target functions are in the same module as the mutator function.

        :param mutating_fn: Function to validate against
        :return: Nothing, raises exception if not valid.
        """
        local_module = mutating_fn.__module__
        for remote_applicable in self.remote_applicables:
            if remote_applicable.target_fn.__module__ != local_module:
                raise NotSameModuleError(fn=mutating_fn, target_fn=remote_applicable.target_fn)

    def _create_step(self, mutating_fn: Callable, remote_applicable_builder: Applicable):
        """Adds the correct function for the applicable and resolves kwargs"""

        if not remote_applicable_builder.kwargs:
            remote_applicable_builder.kwargs = self.mutating_function_kwargs

        remote_applicable_builder.fn = mutating_fn

        return remote_applicable_builder

    def __call__(self, mutating_fn: Callable):
        """Adds to an existing pipe_output or creates a new pipe_output.

        This is a new type of decorator that builds ``pipe_output`` for multiple nodes in the DAG. It does
        not fit in the current decorator framework since it does not decorate the node function in the DAG
        but allows us to "remotely decorate" multiple nodes at once, which needs to happen before the
        NodeTransformLifecycle gets applied / resolved.

        :param mutating_fn: function that will be used in pipe_output to transform target function
        :return: mutating_fn, to guarantee function works even when Hamilton driver is not used
        """

        # This function will be excluded from the DAG as a node since we are inserting it manually
        mutating_fn = hamilton_exclude()(mutating_fn)

        if self.restrict_to_single_module:
            self.validate_same_module(mutating_fn=mutating_fn)

        # TODO:  If @mutate runs once it's good
        #       If you run that again, it might double-up
        # In the juptyer notebook/cross-module case we'll want to guard against it.
        for remote_applicable in self.remote_applicables:
            new_pipe_step = self._create_step(
                mutating_fn=mutating_fn, remote_applicable_builder=remote_applicable
            )
            found_pipe_output = False
            if hasattr(remote_applicable.target_fn, base.NodeTransformer.get_lifecycle_name()):
                for decorator in remote_applicable.target_fn.transform:
                    if isinstance(decorator, pipe_output):
                        decorator.transforms = decorator.transforms + (new_pipe_step,)
                        found_pipe_output = True

            if not found_pipe_output:
                remote_applicable.target_fn = pipe_output(
                    new_pipe_step, collapse=self.collapse, _chain=self.chain
                )(remote_applicable.target_fn)

        return mutating_fn



from typing import Callable

import pytest

from hamilton import settings
from hamilton.function_modifiers import (
    ResolveAt,
    base,
    extract_columns,
    resolve,
    resolve_from_config,
)

CONFIG_WITH_POWER_MODE_ENABLED = {
    settings.ENABLE_POWER_USER_MODE: True,
}

CONFIG_WITH_POWER_MODE_DISABLED = {
    settings.ENABLE_POWER_USER_MODE: False,
}


@pytest.mark.parametrize(
    "fn,required,optional",
    [
        (lambda: 1, [], {}),
        (lambda a, b: 1, ["a", "b"], {}),
        (lambda a, b=1: 1, ["a"], {"b": 1}),
        (lambda a=1, b=1: 1, [], {"a": 1, "b": 1}),
    ],
)
def test_extract_and_validate_params_happy(fn: Callable, required: Callable, optional: Callable):
    from hamilton.function_modifiers import delayed

    assert delayed.extract_and_validate_params(fn) == (required, optional)


@pytest.mark.parametrize(
    "fn",
    [
        lambda **kwargs: 1,
        lambda a, b, *args: 1,
        lambda a, b, *args, **kwargs: 1,
    ],
)
def test_extract_and_validate_params_unhappy(fn: Callable):
    from hamilton.function_modifiers import delayed

    with pytest.raises(base.InvalidDecoratorException):
        delayed.extract_and_validate_params(fn)


def test_dynamic_resolves():
    decorator = resolve(
        when=ResolveAt.CONFIG_AVAILABLE,
        decorate_with=lambda cols_to_extract: extract_columns(*cols_to_extract),
    )
    decorator_resolved = decorator.resolve(
        {"cols_to_extract": ["a", "b"], **CONFIG_WITH_POWER_MODE_ENABLED}, fn=test_dynamic_resolves
    )
    # This uses an internal component of extract_columns
    # We may want to add a little more comprehensive testing
    # But for now this will work
    assert decorator_resolved.columns == ("a", "b")


def test_dynamic_resolve_with_configs():
    decorator = resolve_from_config(
        decorate_with=lambda cols_to_extract: extract_columns(*cols_to_extract),
    )
    decorator_resolved = decorator.resolve(
        {"cols_to_extract": ["a", "b"], **CONFIG_WITH_POWER_MODE_ENABLED}, fn=test_dynamic_resolves
    )
    # This uses an internal component of extract_columns
    # We may want to add a little more comprehensive testing
    # But for now this will work
    assert decorator_resolved.columns == ("a", "b")


def test_dynamic_fails_without_power_mode_fails():
    decorator = resolve(
        when=ResolveAt.CONFIG_AVAILABLE,
        decorate_with=lambda cols_to_extract: extract_columns(*cols_to_extract),
    )
    with pytest.raises(base.InvalidDecoratorException):
        decorator_resolved = decorator.resolve(
            CONFIG_WITH_POWER_MODE_DISABLED, fn=test_dynamic_fails_without_power_mode_fails
        )
        # This uses an internal component of extract_columns
        # We may want to add a little more comprehensive testing
        # But for now this will work
        assert decorator_resolved.columns == ("a", "b")


def test_config_derivation():
    decorator = resolve(
        when=ResolveAt.CONFIG_AVAILABLE,
        decorate_with=lambda cols_to_extract,
        some_cols_you_might_want_to_extract=[]: extract_columns(
            *cols_to_extract + some_cols_you_might_want_to_extract
        ),
    )
    assert decorator.required_config() == ["cols_to_extract"]
    assert decorator.optional_config() == {
        "some_cols_you_might_want_to_extract": [],
    }


def test_delayed_with_optional():
    decorator = resolve(
        when=ResolveAt.CONFIG_AVAILABLE,
        decorate_with=lambda cols_to_extract,
        some_cols_you_might_want_to_extract=["c"]: extract_columns(
            *cols_to_extract + some_cols_you_might_want_to_extract
        ),
    )
    resolved = decorator.resolve(
        {"cols_to_extract": ["a", "b"], **CONFIG_WITH_POWER_MODE_ENABLED},
        fn=test_delayed_with_optional,
    )
    assert list(resolved.columns) == ["a", "b", "c"]
    resolved = decorator.resolve(
        {
            "cols_to_extract": ["a", "b"],
            "some_cols_you_might_want_to_extract": ["d"],
            **CONFIG_WITH_POWER_MODE_ENABLED,
        },
        fn=test_delayed_with_optional,
    )
    assert list(resolved.columns) == ["a", "b", "d"]


def test_delayed_without_power_mode_fails():
    decorator = resolve(
        when=ResolveAt.CONFIG_AVAILABLE,
        decorate_with=lambda cols_to_extract,
        some_cols_you_might_want_to_extract=["c"]: extract_columns(
            *cols_to_extract + some_cols_you_might_want_to_extract
        ),
    )
    with pytest.raises(base.InvalidDecoratorException):
        decorator.resolve(
            {"cols_to_extract": ["a", "b"], **CONFIG_WITH_POWER_MODE_DISABLED},
            fn=test_delayed_with_optional,
        )



import functools
import itertools
import json
import os
from typing import Union

import pandas as pd
from fastapi import FastAPI
from fastapi.responses import HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastui import AnyComponent, FastUI, prebuilt_html
from fastui import components as c
from fastui.components.display import DisplayLookup, DisplayMode
from fastui.events import GoToEvent
from fastui.forms import SelectSearchResponse
from pydantic import BaseModel, Field

from hamilton.plugins.h_experiments.cache import JsonCache
from hamilton.plugins.h_experiments.data_model import (
    NodeMaterializer,
    RunMetadata,
    model_from_values,
)


def convert_graph_hash_to_version(runs):
    """Convert graph hash to incremental id, per experiment"""
    versioned_runs = []
    for _, group in itertools.groupby(runs, lambda r: r.experiment):
        group = sorted(group, key=lambda run: run.date_completed)

        current_version = 1
        current_hash = None
        for r in group:
            if current_hash is None:
                current_hash = r.graph_hash

            if r.graph_hash != current_hash:
                current_hash = r.graph_hash
                current_version += 1

            r.graph_version = current_version
            versioned_runs.append(r)
    return versioned_runs


def get_runs(metadata_cache_path: str) -> list[RunMetadata]:
    """Create RunMetadata objects for all runs JSON found in cache"""
    cache = JsonCache(cache_path=metadata_cache_path)
    runs = [RunMetadata.model_validate_json(cache.read(run_id)) for run_id in cache.keys()]
    runs = convert_graph_hash_to_version(runs)
    runs = sorted(runs, key=lambda run: run.date_completed, reverse=True)
    return runs


# environment variables are the most convenient approach
# to configure FastAPI application
base_directory = os.getenv("HAMILTON_EXPERIMENTS_PATH", "")

# disable Swagger UI /docs because they are currently bugged for FastUI
app = FastAPI(docs_url=None, redoc_url=None)
app.mount("/experiments", StaticFiles(directory=base_directory), name="experiments")
runs = get_runs(base_directory)


def base_page(*components: AnyComponent) -> list[AnyComponent]:
    """Template applied to all pages. Includes: title, navigation, and footer"""
    return [
        c.PageTitle(text="📝 Hamilton Experiment Manager"),
        c.Navbar(
            title="📝 Hamilton Experiment Manager",
            title_event=GoToEvent(url="/"),
            class_name="+ mb-4",
        ),
        c.Page(components=[*components]),
        c.Footer(
            extra_text="Powered by Hamilton",
            links=[
                c.Link(
                    components=[c.Text(text="GitHub")],
                    on_click=GoToEvent(url="https://github.com/dagworks-inc/hamilton"),
                ),
            ],
        ),
    ]


@functools.cache
def run_lookup():
    """Cache a mapping of {run_id: run} to query runs"""
    return {run.run_id: run for run in runs}


@app.get("/api/filter/{field}", response_model=SelectSearchResponse)
async def search_filter(field: str) -> SelectSearchResponse:
    """Get all the unique values for a RunMetadata field to populate menus"""
    options = {str(getattr(run, field)): getattr(run, field) for run in runs}
    return SelectSearchResponse(options=[{"label": v, "value": v} for v in options])


class RunFilter(BaseModel):
    """Filter the runs_overview() page using `experiment` and `graph_version`"""

    experiment: str = Field(
        json_schema_extra={
            "search_url": "/api/filter/experiment",
            "placeholder": "Select Experiment ...",
        }
    )
    graph_version: str = Field(
        json_schema_extra={
            "search_url": "/api/filter/graph_version",
            "placeholder": "Select Graph version ...",
        },
    )


@app.get("/api/runs", response_model=FastUI, response_model_exclude_none=True)
def runs_overview(
    experiment: Union[str, None] = None,
    graph_version: Union[int, None] = None,
) -> list[AnyComponent]:
    """RunOverview page with filters for the table"""

    # refresh cache to display new runs
    runs = get_runs(os.getenv("HAMILTON_EXPERIMENTS_PATH", ""))

    selection = dict()

    if experiment:
        selection["experiment"] = dict(value=experiment, label=experiment)

    if graph_version:
        selection["graph_version"] = dict(value=graph_version, label=str(graph_version))

    selected_runs = runs
    if selection:
        for k, v in selection.items():
            selected_runs = list(filter(lambda r: getattr(r, k) == v["value"], selected_runs))

    return base_page(
        c.ModelForm(
            model=RunFilter,
            submit_url="/runs",
            initial=selection,
            method="GOTO",
            submit_on_change=True,
            display_mode="inline",
        ),
        c.Table(
            data=selected_runs,
            data_model=RunMetadata,
            columns=[
                DisplayLookup(field="date_completed", mode=DisplayMode.datetime),
                DisplayLookup(field="experiment"),
                DisplayLookup(field="graph_version"),
                DisplayLookup(field="run_id", on_click=GoToEvent(url="/run/{run_id}/")),
            ],
        ),
    )


def run_tabs(run_id) -> list[AnyComponent]:
    """Create Metadata and Artifacts tabs for individual Run pages"""
    return [
        c.LinkList(
            links=[
                c.Link(
                    components=[c.Text(text="Metadata")],
                    on_click=GoToEvent(url=f"/run/{run_id}/"),
                    active="startswith:/run/",
                ),
                c.Link(
                    components=[c.Text(text="Artifacts")],
                    on_click=GoToEvent(url=f"/artifacts/{run_id}"),
                    active="startswith:/artifacts/",
                ),
            ],
            mode="tabs",
            class_name="+ mb-4",
        )
    ]


@app.get("/api/run/{run_id}/", response_model=FastUI, response_model_exclude_none=True)
def run_metadata(run_id: str) -> list[AnyComponent]:
    """Individual Run > Metadata"""
    run = run_lookup()[run_id]

    return base_page(
        c.Heading(text=run.experiment, level=2),
        *run_tabs(run_id=run_id),
        c.Details(
            data=run,
            fields=[
                DisplayLookup(field="experiment"),
                DisplayLookup(field="run_id"),
                DisplayLookup(field="success"),
                DisplayLookup(field="graph_hash"),
                DisplayLookup(field="modules"),
            ],
        ),
        c.Image(
            src=f"/experiments/{run.experiment}/{run.run_id}/dag.png",
            width="100%",
            height="auto",
            loading="lazy",
            referrer_policy="no-referrer",
            class_name="border rounded",
        ),
        c.Details(
            data=run,
            fields=[
                DisplayLookup(field="config", mode=DisplayMode.json),
                DisplayLookup(field="inputs", mode=DisplayMode.json),
                DisplayLookup(field="overrides", mode=DisplayMode.json),
            ],
        ),
    )


@functools.cache
def create_table_model(**kwargs):
    """Cache the creation of a Pydantic model from a DataFrame row"""
    return model_from_values("Table", specs=kwargs)


def dataframe_to_table(df: pd.DataFrame, **kwargs) -> AnyComponent:
    """Populate a FastUI table with pagination from a pandas DataFrame"""
    df = df.reset_index()
    Table = create_table_model(**{str(k): v for k, v in df.iloc[0].to_dict().items()})

    page: int = 1 if kwargs.get("page") is None else kwargs.get("page", 1)
    page_size = min(20, df.shape[0])
    page_df = df.iloc[(page - 1) * page_size : page * page_size]

    # create a Pydantic objects for each row
    data = [
        Table.model_validate({str(k): v for k, v in row.items()})
        for row in page_df.to_dict(orient="index").values()
    ]

    return [
        c.Table(
            data=data,
            data_model=Table,
            columns=[DisplayLookup(field=str(col)) for col in df.columns],
        ),
        c.Pagination(page=page, page_size=20, total=df.shape[0]),
    ]


def artifact_components(materializer: NodeMaterializer, **kwargs) -> list[AnyComponent]:
    """Create a FastUI component for an artifact based on the materializer type

    Instead of mapping Python type -> component, the materializer/file type -> component
    has a much lower and manageable cardinality
    """
    artifact_path = materializer.path.replace(base_directory, "/experiments")

    if materializer.sink == "json":
        with open(materializer.path, "r") as f:
            data = json.load(f)
        components = [c.Json(value=data)]

    elif materializer.sink == "parquet":
        data = pd.read_parquet(materializer.path)
        components = dataframe_to_table(data, **kwargs)

    elif materializer.sink == "csv":
        data = pd.read_csv(materializer.path)
        components = dataframe_to_table(data, **kwargs)

    elif materializer.sink in ["plt", "plotly"]:
        try:
            components = [
                c.Image(
                    src=artifact_path,
                    width="100%",
                    height="auto",
                    loading="lazy",
                    referrer_policy="no-referrer",
                )
            ]
        # TODO refactor to each default else case
        except Exception:
            components = [
                c.Json(
                    value={
                        k: v
                        for k, v in dict(materializer).items()
                        if k in ["path", "sink", "data_saver"]
                    }
                )
            ]

    else:
        components = [
            c.Json(
                value={
                    k: v
                    for k, v in dict(materializer).items()
                    if k in ["path", "sink", "data_saver"]
                }
            )
        ]

    return components


def artifact_tabs(run: RunMetadata) -> list[AnyComponent]:
    """Create a tab for each artifact"""
    links = []
    for i, m in enumerate(run.materialized):
        artifact_name = "-".join(m.source_nodes)
        artifact_name, _, _ = artifact_name.partition("__")
        links.append(
            c.Link(
                components=[c.Text(text=artifact_name)],
                on_click=GoToEvent(
                    url=f"/artifacts/{run.run_id}",
                    query=dict(artifact_id=i),
                ),
                active=f"startswith:/artifacts/{run.run_id}?artifact_id={i}",
            )
        )

    return [
        c.LinkList(
            links=links,
            mode="tabs",
            class_name="+ mb-4",
        )
    ]


@app.get("/api/artifacts/{run_id}", response_model=FastUI, response_model_exclude_none=True)
def run_artifacts(
    run_id: str, artifact_id: int = 0, page: Union[int, None] = None
) -> list[AnyComponent]:
    """Individual Run > Artifact"""
    run = run_lookup()[run_id]

    return base_page(
        c.Heading(text=run.experiment, level=2),
        *run_tabs(run_id=run_id),
        *artifact_tabs(run=run),
        *artifact_components(materializer=run.materialized[artifact_id], page=page),
    )


@app.get("/api/")
def landing_page() -> RedirectResponse:
    """Landing page redirects to the Run overview page"""
    return RedirectResponse(url="/api/runs")


@app.get("{path:path}")
async def html_landing() -> HTMLResponse:
    """Simple HTML page which serves the React app, comes last as it matches all paths."""
    return HTMLResponse(prebuilt_html(title="Hamilton Experiment Manager"))



import logging

from hamilton.plugins.h_spark import (
    KoalasDataFrameResult,
    PySparkUDFGraphAdapter,
    SparkKoalasGraphAdapter,
)

logger = logging.getLogger(__name__)
logger.warning(
    "Importing from this module is deprecated. We have moved these features out of experimental!"
    " Please use hamilton.plugins.h_spark instead."
)

__all__ = ["KoalasDataFrameResult", "SparkKoalasGraphAdapter", "PySparkUDFGraphAdapter"]



"""Decorators that attach metadata to nodes"""

import json
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

from hamilton import htypes, node, registry
from hamilton.function_modifiers import base

RAY_REMOTE_TAG_NAMESPACE = "ray_remote"


class tag(base.NodeDecorator):
    """Decorator class that adds a tag to a node. Tags take the form of key/value pairings.
    Tags can have dots to specify namespaces (keys with dots), but this is usually reserved for special cases
    (E.G. subdecorators) that utilize them. Usually one will pass in tags as kwargs, so we expect tags to
    be un-namespaced in most uses.

    That is using:

    .. code-block:: python

       @tag(my_tag='tag_value')
       def my_function(...) -> ...:

    is un-namespaced because you cannot put a `.` in the keyword part (the part before the '=').

    But using:

    .. code-block:: python

       @tag(**{'my.tag': 'tag_value'})
       def my_function(...) -> ...:

    allows you to add dots that allow you to namespace your tags.

    Currently, tag values are restricted to allowing strings only, although we may consider changing the in the future
    (E.G. thinking of lists).

    Hamilton also reserves the right to change the following:
    * adding purely positional arguments
    * not allowing users to use a certain set of top-level prefixes (E.G. any tag where the top level is one of the \
    values in RESERVED_TAG_PREFIX).

    Example usage:

    .. code-block:: python

       @tag(foo='bar', a_tag_key='a_tag_value', **{'namespace.tag_key': 'tag_value'})
       def my_function(...) -> ...:
          ...
    """

    RESERVED_TAG_NAMESPACES = [
        "hamilton",
        "data_quality",
        "gdpr",
        "ccpa",
        "dag",
        "module",
        RAY_REMOTE_TAG_NAMESPACE,
    ]  # Anything that starts with any of these is banned, the framework reserves the right to manage it

    def __init__(
        self,
        *,
        target_: base.TargetType = None,
        bypass_reserved_namespaces_: bool = False,
        **tags: Union[str, List[str]],
    ):
        """Constructor for adding tag annotations to a function.

        :param bypass_reserved_namespaces\\_: Whether to bypass Reserved Namespace checking.
        :param target\\_: Target nodes to decorate. This can be one of the following:

            * **None**: tag all nodes outputted by this that are "final" (E.g. do not have a node\
            outputted by this that depend on them)
            * **Ellipsis (...)**: tag *all* nodes outputted by this
            * **Collection[str]**: tag *only* the nodes with the specified names
            * **str**: tag *only* the node with the specified name
        :param tags: the keys are always going to be strings, so the type annotation here means the values are strings \
            or lists of values. Implicitly this is `Dict[str, Union[str, List[str]]]` but the PEP guideline is to only
             annotate it with the value `Union[str, List[str]]`.
        """
        super(tag, self).__init__(target=target_)
        self.tags = tags
        self.bypass_reserved_namespaces = bypass_reserved_namespaces_

    def decorate_node(self, node_: node.Node) -> node.Node:
        """Decorates the nodes produced by this with the specified tags

        :param node_: Node to decorate
        :return: Copy of the node, with tags assigned
        """
        node_tags = node_.tags.copy()
        node_tags.update(self.tags)
        return node_.copy_with(tags=node_tags)

    def _key_allowed(self, key: str) -> bool:
        """Validates that a tag key is allowed. Rules are:
        1. It must not be empty
        2. It can have dots, which specify a hierarchy of order
        3. All components, when split by dots, must be valid python identifiers
        4. It cannot utilize a reserved namespace

        :param key: The key to validate
        :return: True if it is valid, False if not
        """
        key_components = key.split(".")
        if len(key_components) == 0:
            # empty string...
            return False
        if not self.bypass_reserved_namespaces and key_components[0] in tag.RESERVED_TAG_NAMESPACES:
            # Reserved prefixes
            return False
        for key in key_components:
            if not key.isidentifier():
                return False
        return True

    @staticmethod
    def _value_allowed(value: Any) -> bool:
        """Validates that a tag value is allowed. Rules are only that it must be a string.

        :param value: Value to validate
        :return: True if it is valid, False otherwise
        """
        if not isinstance(value, str) and not isinstance(value, list):
            return False
        elif isinstance(value, list):
            if len(value) == 0:
                return False  # disallow empty lists
            if not all([isinstance(v, str) for v in value]):  # need values to be all strings
                return False
            if len(set(value)) != len(value):  # need values to be unique
                return False
        return True

    def validate(self, fn: Callable):
        """Validates the decorator. In this case that the set of tags produced is final.

        :param fn: Function that the decorator is called on.
        :raises ValueError: if the specified tags contains invalid ones
        """
        bad_tags = set()
        for key, value in self.tags.items():
            if (not self._key_allowed(key)) or (not tag._value_allowed(value)):
                if isinstance(value, list):
                    value = str(value)
                bad_tags.add((key, value))
        if bad_tags:
            bad_tags_formatted = ",".join([f"{key}={value}" for key, value in bad_tags])
            raise base.InvalidDecoratorException(
                f"The following tags are invalid as tags: {bad_tags_formatted} "
                "Tag keys can be split by ., to represent a hierarchy, "
                "but each element of the hierarchy must be a valid python identifier. "
                "Paths components also cannot be empty. "
                "The value can only be a string, or a list of strings. "
                "Note that the following top-level prefixes are "
                f"reserved as well: {self.RESERVED_TAG_NAMESPACES}"
            )


class tag_outputs(base.NodeDecorator):
    def __init__(self, **tag_mapping: Dict[str, Union[str, List[str]]]):
        """Creates a tag_outputs decorator.

        Note that this currently does not validate whether the nodes are spelled correctly as it takes in a superset of\
        nodes.

        :param tag_mapping: Mapping of output name to tags -- this is akin to applying @tag to individual outputs \
        produced by the function.

        Example usage:

        .. code-block:: python

           @tag_output(**{'a': {'a_tag': 'a_tag_value'}, 'b': {'b_tag': 'b_tag_value'}})
           @extract_columns("a", "b")
           def example_tag_outputs() -> pd.DataFrame:
               return pd.DataFrame.from_records({"a": [1], "b": [2]})

        """
        super(base.NodeDecorator, self).__init__(target=...)
        self.tag_mapping = tag_mapping

    def decorate_node(self, node_: node.Node) -> node.Node:
        """Decorates all final nodes with the specified tags."""
        if node_.name not in self.tag_mapping:
            return node_  # in this case we have no desire to update tags
        new_tags = node_.tags.copy()
        new_tags.update(self.tag_mapping.get(node_.name, {}))
        return tag(**new_tags).decorate_node(node_)


# These represent a generic schema type -- E.G. one that will
# be supported across the entire set of usable dataframe/dataset types
# Eventually we'll be integrating mappings of these into the registry,
# but for now this serves largely as a placeholder/documentation
# GENERIC_SCHEMA_TYPES = (
#     "int",
#     "float",
#     "str",
#     "bool",
#     "dict",
#     "list",
#     "object",
#     "datetime",
#     "date",
# )


class SchemaOutput(tag):
    def __init__(self, *fields: Tuple[str, str], target_: Optional[str] = None):
        """Initializes SchemaOutput. See docs for `@schema.output` for more details."""

        tag_value = ",".join([f"{key}={value}" for key, value in fields])
        super(SchemaOutput, self).__init__(
            **{schema.INTERNAL_SCHEMA_OUTPUT_KEY: tag_value}, target_=target_
        )

    def validate_node(self, node_: node.Node):
        """Validates that the node has a return type of a registered dataframe.

        :param node_: Node to validate
        :raises InvalidDecoratorException: if the node does not have a return type of a registered dataframe.
        """
        output_type = node_.type
        available_types = registry.get_registered_dataframe_types()
        for _, type_ in available_types.items():
            if htypes.custom_subclass_check(output_type, type_):
                return
        raise base.InvalidDecoratorException(
            f"Node {node_.name} has type {output_type} which is not a registered type for a dataset. "
            f"Registered types are {available_types}. If you found this, either (a) ensure you have the "
            f"right package installed, or (b) reach out to the team to figure out how to add yours."
        )

    @classmethod
    def allows_multiple(cls) -> bool:
        """Currently this only applies to a single output. If it is a set of nodes with multiple outputs,
        it will apply to the "final" (sink) one. We can change this if there's need."""
        return False

    def validate(self, fn: Callable):
        """Bypassed for now -- we have no function-level or class-level validations yet,
        but this is done at `@tag`, which this inherits. We will be moving away from inheriting tag.
        """
        pass


class schema:
    """Container class for schema stuff. This is purely so we can have a nice API for it -- E.G. Schema.output"""

    INTERNAL_SCHEMA_OUTPUT_KEY = "hamilton.internal.schema_output"

    @staticmethod
    def output(*fields: Tuple[str, str], target_: Optional[str] = None) -> SchemaOutput:
        """Initializes a `@schema.output` decorator. This takes in a list of fields, which are tuples of the form
        `(field_name, field_type)`. The field type must be one of the function_modifiers.SchemaTypes types.

        :param target_: Target node to decorate -- if `None` it'll decorate all final nodes (E.G. sinks in the subdag),
            otherwise it will decorate the specified node.
        :param fields: List of fields to add to the schema. Each field is a tuple of the form `(field_name, field_type)`

        This is implemented using tags, but that might change. Thus you should not
        rely on the tags created by this decorator (which is why they are prefixed with `internal`).

        To use this, you should decorate a node with `@schema.output`

        Example usage:

        .. code-block:: python

           @schema.output(
               ("a", "int"),
               ("b", "float"),
               ("c", "str")
            )
           def example_schema() -> pd.DataFrame:
               return pd.DataFrame.from_records({"a": [1], "b": [2.0], "c": ["3"]})

        Then, when drawing the DAG, the schema will be displayed as sub-elements in the node for the DAG (if `display_schema` is selected).
        """
        return SchemaOutput(*fields, target_=target_)


class RayRemote(tag):
    def __init__(self, **options: Union[int, Dict[str, int]]):
        """Initializes RayRemote. See docs for `@ray_remote_options` for more details."""

        ray_tags = {f"ray_remote.{option}": json.dumps(value) for option, value in options.items()}

        super(RayRemote, self).__init__(bypass_reserved_namespaces_=True, **ray_tags)


def ray_remote_options(**kwargs: Union[int, Dict[str, int]]) -> RayRemote:
    """Initializes a `@ray_remote_options` decorator. This takes in a list of options to pass to ray.remote().

    Supported options include resources, as well as other options:
    https://docs.ray.io/en/latest/ray-core/scheduling/resources.html

    This is implemented using tags, but that might change. Thus you should not
    rely on the tags created by this decorator (which is why they are on a reserved namespace).

    To use this, you should decorate a node with `@ray_remote_options`

    Example usage:

    .. code-block:: python

        @ray_remote_options(
            num_gpus=1,
            resources={"my_custom_resource": 1},
        )
        def example() -> pd.DataFrame: ...
    """
    return RayRemote(**kwargs)


# materializers that have a `path` kwarg and are part of the core Hamilton library
# parquet, csv, feather, orc, and excel are via the pandas extension because it's currently a Hamilton dependency
CACHE_MATERIALIZERS = Literal[
    "json",
    "file",
    "pickle",
    "parquet",
    "csv",
    "feather",
    "orc",
    "excel",
]

# see hamilton.caching.adapter.CachingBehavior enum for details.
# default: caching is enabled
# recompute: always compute the node instead of retrieving
# ignore: the data version won't be part of downstream keys
# disable: act as if caching wasn't enabled.
CACHE_BEHAVIORS = Literal["default", "recompute", "ignore", "disable"]


class cache(base.NodeDecorator):
    BEHAVIOR_KEY = "cache.behavior"
    FORMAT_KEY = "cache.format"

    def __init__(
        self,
        *,
        behavior: Optional[CACHE_BEHAVIORS] = None,
        format: Optional[Union[CACHE_MATERIALIZERS, str]] = None,
        target_: base.TargetType = ...,
    ):
        """The ``@cache`` decorator can define the behavior and format of a specific node.

        This feature is implemented via tags, but that could change. Thus you should not
        rely on these tags for other purposes.

        .. code-block:: python

            @cache(behavior="recompute", format="parquet")
            def raw_data() -> pd.DataFrame: ...


        If the function uses other function modifiers and define multiple nodes, you can
        set ``target_`` to specify which nodes to cache. The following only caches the ``performance`` node.

        .. code-block:: python

            @cache(format="json", target_="performance")
            @extract_fields(trained_model=LinearRegression, performance: dict)
            def model_training() -> dict:
                # ...
                performance = {"rmse": 0.1, "mae": 0.2}
                return {"trained_model": trained_model, "performance": performance}


        :param behavior: The behavior of the cache. This can be one of the following:
            * **default**: caching is enabled
            * **recompute**: always compute the node instead of retrieving
            * **ignore**: the data version won't be part of downstream keys
            * **disable**: act as if caching wasn't enabled.
        :param format: The format of the cache. This can be one of the following:
            * **json**: JSON format
            * **file**: file format
            * **pickle**: pickle format
            * **parquet**: parquet format
            * **csv**: csv format
            * **feather**: feather format
            * **orc**: orc format
            * **excel**: excel format
        :param target\\_: Target nodes to decorate. This can be one of the following:
            * **None**: tag all nodes outputted by this that are "final" (E.g. do not have a node\
            outputted by this that depend on them)
            * **Ellipsis (...)**: tag *all* nodes outputted by this
            * **Collection[str]**: tag *only* the nodes with the specified names
            * **str**: tag *only* the node with the specified name
        """
        super(cache, self).__init__(target=target_)

        # don't provide default value for behavior and format if not provided by user
        # the SmartCacheAdapter expects the field to be empty if not set
        self.cache_tags = {}
        if behavior:
            self.cache_tags[cache.BEHAVIOR_KEY] = behavior

        if format:
            self.cache_tags[cache.FORMAT_KEY] = format

    def decorate_node(self, node_: node.Node) -> node.Node:
        """Decorates the nodes with the cache tags.

        :param node_: Node to decorate
        :return: Copy of the node, with tags assigned
        """
        node_tags = node_.tags.copy()
        node_tags.update(self.cache_tags)
        return node_.copy_with(tags=node_tags)



import pathlib
import sqlite3
import threading
from typing import List, Optional

from hamilton.caching.cache_key import decode_key
from hamilton.caching.stores.base import MetadataStore


class SQLiteMetadataStore(MetadataStore):
    def __init__(
        self,
        path: str,
        connection_kwargs: Optional[dict] = None,
    ) -> None:
        self._directory = pathlib.Path(path).resolve()
        self._directory.mkdir(parents=True, exist_ok=True)
        self._path = self._directory.joinpath("metadata_store").with_suffix(".db")
        self.connection_kwargs: dict = connection_kwargs if connection_kwargs else {}

        self._thread_local = threading.local()

        # creating tables at `__init__` prevents other methods from encountering
        # `sqlite3.OperationalError` because tables are missing.
        self._create_tables_if_not_exists()

    def __getstate__(self) -> dict:
        """Serialized `__init__` arguments required to initialize the
        MetadataStore in a new thread or process.
        """
        state = {}
        # NOTE kwarg `path` is not equivalent to `self._path`
        state["path"] = self._directory
        state["connection_kwargs"] = self.connection_kwargs
        return state

    def _get_connection(self) -> sqlite3.Connection:
        if not hasattr(self._thread_local, "connection"):
            self._thread_local.connection = sqlite3.connect(
                str(self._path), check_same_thread=False, **self.connection_kwargs
            )
        return self._thread_local.connection

    def _close_connection(self) -> None:
        if hasattr(self._thread_local, "connection"):
            self._thread_local.connection.close()
            del self._thread_local.connection

    @property
    def connection(self) -> sqlite3.Connection:
        """Connection to the SQLite database."""
        return self._get_connection()

    def __del__(self):
        """Close the SQLite connection when the object is deleted"""
        self._close_connection()

    def _create_tables_if_not_exists(self) -> None:
        """Create the tables necessary for the cache:

        run_ids: queue of run_ids, ordered by start time.
        history: queue of executed node; allows to query "latest" execution of a node
        cache_metadata: information to determine if a node needs to be computed or not

        In the table ``cache_metadata``, the ``cache_key`` is unique whereas
        ``history`` allows duplicate.
        """
        cur = self.connection.cursor()

        cur.execute(
            """\
            CREATE TABLE IF NOT EXISTS run_ids (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                run_id TEXT,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
            )
            """
        )
        cur.execute(
            """\
            CREATE TABLE IF NOT EXISTS history (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                cache_key TEXT,
                run_id TEXT,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,

                FOREIGN KEY (cache_key) REFERENCES cache_metadata(cache_key)
            );
            """
        )
        cur.execute(
            """\
            CREATE TABLE IF NOT EXISTS cache_metadata (
                cache_key TEXT PRIMARY KEY,
                data_version TEXT NOT NULL,
                node_name TEXT NOT NULL,
                code_version TEXT NOT NULL,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,

                FOREIGN KEY (cache_key) REFERENCES history(cache_key)
            );
            """
        )
        self.connection.commit()

    def initialize(self, run_id) -> None:
        """Call initialize when starting a run. This will create database tables
        if necessary.
        """
        cur = self.connection.cursor()
        cur.execute("INSERT INTO run_ids (run_id) VALUES (?)", (run_id,))
        self.connection.commit()

    def __len__(self) -> int:
        """Number of entries in cache_metadata"""
        cur = self.connection.cursor()
        cur.execute("SELECT COUNT(*) FROM cache_metadata")
        return cur.fetchone()[0]

    def set(
        self,
        *,
        cache_key: str,
        data_version: str,
        run_id: str,
        node_name: str = None,
        code_version: str = None,
        **kwargs,
    ) -> None:
        cur = self.connection.cursor()

        # if the caller of ``.set()`` directly provides the ``node_name`` and ``code_version``,
        # we can skip the decoding step.
        if (node_name is None) or (code_version is None):
            try:
                decoded_key = decode_key(cache_key)
            except BaseException as e:
                raise ValueError(
                    f"Failed decoding the cache_key: {cache_key}.\n",
                    "The `cache_key` must be created by `hamilton.caching.cache_key.create_cache_key()` ",
                    "if `node_name` and `code_version` are not provided.",
                ) from e

            node_name = decoded_key["node_name"]
            code_version = decoded_key["code_version"]

        cur.execute("INSERT INTO history (cache_key, run_id) VALUES (?, ?)", (cache_key, run_id))
        cur.execute(
            """\
            INSERT OR IGNORE INTO cache_metadata (
                cache_key, node_name, code_version, data_version
            ) VALUES (?, ?, ?, ?)
            """,
            (cache_key, node_name, code_version, data_version),
        )

        self.connection.commit()

    def get(self, cache_key: str) -> Optional[str]:
        cur = self.connection.cursor()
        cur.execute(
            """\
            SELECT data_version
            FROM cache_metadata
            WHERE cache_key = ?
            """,
            (cache_key,),
        )
        result = cur.fetchone()

        if result is None:
            data_version = None
        else:
            data_version = result[0]

        return data_version

    def delete(self, cache_key: str) -> None:
        """Delete metadata associated with ``cache_key``."""
        cur = self.connection.cursor()
        cur.execute("DELETE FROM cache_metadata WHERE cache_key = ?", (cache_key,))
        self.connection.commit()

    def delete_all(self) -> None:
        """Delete all existing tables from the database"""
        cur = self.connection.cursor()

        for table_name in ["run_ids", "history", "cache_metadata"]:
            cur.execute(f"DROP TABLE IF EXISTS {table_name};")

        self.connection.commit()

    def exists(self, cache_key: str) -> bool:
        """boolean check if a ``data_version`` is found for ``cache_key``
        If True, ``.get()`` should successfully retrieve the ``data_version``.
        """
        cur = self.connection.cursor()
        cur.execute("SELECT cache_key FROM cache_metadata WHERE cache_key = ?", (cache_key,))
        result = cur.fetchone()

        return result is not None

    def get_run_ids(self) -> List[str]:
        """Return a list of run ids, sorted from oldest to newest start time."""
        cur = self.connection.cursor()
        cur.execute("SELECT run_id FROM run_ids ORDER BY id")
        result = cur.fetchall()

        return [r[0] for r in result]

    def _run_exists(self, run_id: str) -> bool:
        """Returns True if a run was initialized with ``run_id``, even
        if the run recorded no node executions.
        """
        cur = self.connection.cursor()
        cur.execute(
            """\
            SELECT EXISTS(
                SELECT 1
                FROM run_ids
                WHERE run_id = ?
            )
            """,
            (run_id,),
        )
        result = cur.fetchone()
        # SELECT EXISTS returns 1 for True, i.e., `run_id` is found
        return result[0] == 1

    def get_run(self, run_id: str) -> List[dict]:
        """Return a list of node metadata associated with a run.

        :param run_id: ID of the run to retrieve
        :return: List of node metadata which includes ``cache_key``, ``data_version``,
            ``node_name``, and ``code_version``. The list can be empty if a run was initialized
            but no nodes were executed.

        :raises IndexError: if the ``run_id`` is not found in metadata store.
        """
        cur = self.connection.cursor()
        if self._run_exists(run_id) is False:
            raise IndexError(f"`run_id` not found in table `run_ids`: {run_id}")

        cur.execute(
            """\
            SELECT
                cache_metadata.cache_key,
                cache_metadata.data_version,
                cache_metadata.node_name,
                cache_metadata.code_version
            FROM history
            JOIN cache_metadata ON history.cache_key = cache_metadata.cache_key
            WHERE history.run_id = ?
            """,
            (run_id,),
        )
        results = cur.fetchall()
        return [
            dict(
                cache_key=cache_key,
                data_version=data_version,
                node_name=node_name,
                code_version=code_version,
            )
            for cache_key, data_version, node_name, code_version in results
        ]



import abc
import dataclasses
from typing import Any, Collection, Dict, Tuple, Type

try:
    import pyspark.sql as ps
except ImportError as e:
    raise NotImplementedError("Pyspark is not installed.") from e

from pandas import DataFrame
from pyspark.sql import SparkSession

from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataLoader


@dataclasses.dataclass
class SparkDataFrameDataLoader(DataLoader):
    """Base class for data loaders that load pyspark dataframes.
    We are not yet including data savers, but that will be added to this most likely..
    """

    spark: SparkSession

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [ps.DataFrame]

    @abc.abstractmethod
    def load_data(self, type_: Type[DataFrame]) -> Tuple[ps.DataFrame, Dict[str, Any]]:
        pass


@dataclasses.dataclass
class CSVDataLoader(SparkDataFrameDataLoader):
    path: str  # It supports multiple but for now we're going to have a single one
    # We can always make that a list of strings, or make a multiple reader (.multicsv)
    header: bool = True
    sep: str = ","

    def load_data(self, type_: Type[DataFrame]) -> Tuple[ps.DataFrame, Dict[str, Any]]:
        return (
            self.spark.read.csv(self.path, header=self.header, sep=self.sep, inferSchema=True),
            utils.get_file_metadata(self.path),
        )

    @classmethod
    def name(cls) -> str:
        return "csv"


@dataclasses.dataclass
class ParquetDataLoader(SparkDataFrameDataLoader):
    path: str  # It supports multiple but for now we're going to have a single one

    # We can always make that a list of strings, or make a multiple reader (.multicsv)

    def load_data(self, type_: Type[DataFrame]) -> Tuple[ps.DataFrame, Dict[str, Any]]:
        return self.spark.read.parquet(self.path), utils.get_file_metadata(self.path)

    @classmethod
    def name(cls) -> str:
        return "parquet"


def register_data_loaders():
    """Function to register the data loaders for this extension."""
    for loader in [CSVDataLoader, ParquetDataLoader]:
        registry.register_adapter(loader)


COLUMN_FRIENDLY_DF_TYPE = False

register_data_loaders()


DATAFRAME_TYPE = ps.DataFrame
COLUMN_TYPE = None


def register_types():
    """Function to register the types for this extension."""
    registry.register_types("spark", DATAFRAME_TYPE, COLUMN_TYPE)


register_types()



import abc
import dataclasses
import enum
import logging
from typing import Any, Dict, List, Tuple, Type

logger = logging.getLogger(__name__)


class DataValidationError(Exception):
    pass


class DataValidationLevel(enum.Enum):
    WARN = "warn"
    FAIL = "fail"


@dataclasses.dataclass
class ValidationResult:
    passes: bool  # Whether or not this passed the validation
    message: str  # Error message or success message
    diagnostics: Dict[str, Any] = dataclasses.field(
        default_factory=dict
    )  # Any extra diagnostics information needed, free-form


class DataValidator(abc.ABC):
    """Base class for a data quality operator. This will be used by the `data_quality` operator"""

    def __init__(self, importance: str):
        self._importance = DataValidationLevel(importance)

    @property
    def importance(self) -> DataValidationLevel:
        return self._importance

    @abc.abstractmethod
    def applies_to(self, datatype: Type[Type]) -> bool:
        """Whether or not this data validator can apply to the specified dataset

        :param datatype:
        :return: True if it can be run on the specified type, false otherwise
        """
        pass

    @abc.abstractmethod
    def description(self) -> str:
        """Gives a description of this validator. E.G.
        `Checks whether the entire dataset lies between 0 and 1.`
        Note it should be able to access internal state (E.G. constructor arguments).
        :return: The description of the validator as a string
        """
        pass

    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """Returns the name for this validator."""

    @abc.abstractmethod
    def validate(self, dataset: Any) -> ValidationResult:
        """Actually performs the validation. Note when you

        :param dataset: dataset to validate
        :return: The result of validation
        """
        pass


def act_warn(node_name: str, validation_result: ValidationResult, validator: DataValidator):
    """This is the current default for acting on the validation result when you want to warn.
    Note that we might move this at some point -- we'll want to make it configurable. But for now, this
    seems like a fine place to put it.

    :param node_name: the name of the node we are validating.
    :param validation_result:  the result
    :param validator: the validator object.
    """
    if not validation_result.passes:
        logger.warning(_create_error_string(node_name, validation_result, validator))


def _create_error_string(node_name, validation_result, validator):
    return (
        f"[{node_name}:{validator.name()}] validator failed. Message was: {validation_result.message}. "
        f"Diagnostic information is: {validation_result.diagnostics}."
    )


def act_fail_bulk(node_name: str, failures: List[Tuple[ValidationResult, DataValidator]]):
    """This is the current default for acting on the validation result when you want to fail.
    Note that we might move this at some point -- we'll want to make it configurable. But for now, this
    seems like a fine place to put it.
    :param node_name: the name of the node we are validating.
    :param failures: list of tuples of (validation_result, validator)
    :raises DataValidationError: if there are errors detected.
    """
    error_messages = []
    for validation_result, validator in failures:
        if not validation_result.passes:
            message = f"{_create_error_string(node_name, validation_result, validator)}\n"
            logger.error(message)  # log here so things print nicely at least
            error_messages.append(message)
    if error_messages:
        raise DataValidationError(error_messages)


class BaseDefaultValidator(DataValidator, abc.ABC):
    """Base class for a default validator.
    These are all validators that utilize a single argument to be passed to the decorator check_output.
    check_output can thus delegate to multiple of these. This is an internal abstraction to allow for easy
    creation of validators.
    """

    def __init__(self, importance: str):
        super(BaseDefaultValidator, self).__init__(importance)

    @classmethod
    @abc.abstractmethod
    def applies_to(cls, datatype: Type[Type]) -> bool:
        pass

    @abc.abstractmethod
    def description(self) -> str:
        pass

    @abc.abstractmethod
    def validate(self, data: Any) -> ValidationResult:
        pass

    @classmethod
    @abc.abstractmethod
    def arg(cls) -> str:
        """Yields a string that represents this validator's argument.
        @check_output() will be passed a series of kwargs, each one of which will correspond to
        one of these default validators. Note that we have the limitation of allowing just a single
        argument.

        :return: The argument that this needs.
        """
        pass

    @classmethod
    def name(cls) -> str:
        return f"{cls.arg()}_validator"



import importlib
import json
import math
import sys
from typing import Any, Callable, Dict, List, Type

import pandas as pd
import pytest

from hamilton import ad_hoc_utils, base, driver, settings
from hamilton.base import DefaultAdapter
from hamilton.data_quality.base import DataValidationError, ValidationResult
from hamilton.execution import executors, grouping
from hamilton.function_modifiers import source, value
from hamilton.io.materialization import from_, to

import tests.resources.data_quality
import tests.resources.dynamic_config
import tests.resources.example_module
import tests.resources.overrides
import tests.resources.overrides_from_module
import tests.resources.test_for_materialization


@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: driver.Driver({}, tests.resources.data_quality, adapter=DefaultAdapter())),
        (
            lambda: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.data_quality)
            .with_remote_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .with_local_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .build()
        ),
        (
            lambda: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
            .with_modules(tests.resources.data_quality)
            .build()
        ),
    ],
    ids=["basic_driver", "driver_v2_multithreading", "driver_v2_synchronous"],
)
def test_data_quality_workflow_passes(driver_factory: Callable[[], driver.Driver]):
    dr = driver_factory()
    all_vars = dr.list_available_variables()
    result = dr.execute([var.name for var in all_vars], inputs={"data_quality_should_fail": False})
    dq_nodes = [
        var.name
        for var in all_vars
        if var.tags.get("hamilton.data_quality.contains_dq_results", False)
    ]
    assert len(dq_nodes) == 1
    dq_result = result[dq_nodes[0]]
    assert isinstance(dq_result, ValidationResult)
    assert dq_result.passes is True


@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: driver.Driver({}, tests.resources.data_quality)),
        (
            lambda: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.data_quality)
            .with_remote_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .with_local_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .build()
        ),
        (
            lambda: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
            .with_modules(tests.resources.data_quality)
            .build()
        ),
    ],
    ids=["basic_driver", "driver_v2_multithreading", "driver_v2_synchronous"],
)
def test_data_quality_workflow_fails(driver_factory):
    dr = driver_factory()
    all_vars = dr.list_available_variables()
    with pytest.raises(DataValidationError):
        dr.execute([var.name for var in all_vars], inputs={"data_quality_should_fail": True})


# Adapted from https://stackoverflow.com/questions/41858147/how-to-modify-imported-source-code-on-the-fly
# This is needed to decide whether to import annotations...
def modify_and_import(module_name, package, modification_func):
    spec = importlib.util.find_spec(module_name, package)
    source = spec.loader.get_source(module_name)
    new_source = modification_func(source)
    module = importlib.util.module_from_spec(spec)
    codeobj = compile(new_source, module.__spec__.origin, "exec")
    exec(codeobj, module.__dict__)
    sys.modules[module_name] = module
    return module


@pytest.mark.parametrize(
    "driver_factory,future_import_annotations",
    [
        (lambda modules: driver.Driver({"region": "US"}, *modules), False),
        (
            lambda modules: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(*modules)
            .with_remote_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .with_local_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .with_config({"region": "US"})
            .build(),
            False,
        ),
        (
            lambda modules: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
            .with_config({"region": "US"})
            .with_modules(*modules)
            .build(),
            False,
        ),
        (lambda modules: driver.Driver({"region": "US"}, *modules), True),
        (
            lambda modules: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(*modules)
            .with_remote_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .with_local_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .with_config({"region": "US"})
            .build(),
            True,
        ),
        (
            lambda modules: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
            .with_config({"region": "US"})
            .with_modules(*modules)
            .build(),
            True,
        ),
        (
            lambda modules: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
            .with_grouping_strategy(grouping.GroupNodesAllAsOne())
            .with_config({"region": "US"})
            .with_modules(*modules)
            .build(),
            False,
        ),
        (
            lambda modules: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
            .with_grouping_strategy(grouping.GroupNodesByLevel())
            .with_config({"region": "US"})
            .with_modules(*modules)
            .build(),
            False,
        ),
        (
            lambda modules: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
            .with_grouping_strategy(grouping.GroupNodesIndividually())
            .with_config({"region": "US"})
            .with_modules(*modules)
            .build(),
            False,
        ),
    ],
    ids=[
        "basic_driver_no_future_import",
        "driver_v2_multithreading_no_future_import",
        "driver_v2_synchronous_no_future_import",
        "basic_driver_future_import",
        "driver_v2_multithreading_future_import",
        "driver_v2_synchronous_future_import",
        "driver_v2_group_all_as_one",
        "driver_v2_group_by_level",
        "driver_v2_group_nodes_individually",
    ],
)
def test_smoke_screen_module(driver_factory, future_import_annotations):
    # Monkeypatch the env
    # This tells the smoke screen module whether to use the future import
    modification_func = lambda source: (  # noqa: E731
        "\n".join(["from __future__ import annotations"] + source.splitlines())
        if future_import_annotations
        else source
    )
    module = modify_and_import(
        "tests.resources.smoke_screen_module", tests.resources, modification_func
    )
    dr = driver_factory([module])
    output_columns = [
        "raw_acquisition_cost",
        "pessimistic_net_acquisition_cost",
        "neutral_net_acquisition_cost",
        "optimistic_net_acquisition_cost",
        "series_with_start_date_end_date",
    ]
    res = dr.execute(
        inputs={"date_range": {"start_date": "20200101", "end_date": "20220801"}},
        final_vars=output_columns,
    )
    epsilon = 0.00001
    assert abs(res["raw_acquisition_cost"].mean() - 0.393808) < epsilon
    assert abs(res["pessimistic_net_acquisition_cost"].mean() - 0.420769) < epsilon
    assert abs(res["neutral_net_acquisition_cost"].mean() - 0.405582) < epsilon
    assert abs(res["optimistic_net_acquisition_cost"].mean() - 0.399363) < epsilon
    assert res["series_with_start_date_end_date"].iloc[0] == "date_20200101_date_20220801"


_dynamic_config = {
    "columns_to_sum_map": {
        "ab": ["a", "b"],
        "cd": ["c", "d"],
        "ed": ["e", "d"],
        "ae": ["a", "e"],
        # You can create some crazy self-referential stuff
        "abcd": ["ab", "cd"],
        "abcdcd": ["abcd", "cd"],
    },
    settings.ENABLE_POWER_USER_MODE: True,
}


@pytest.mark.parametrize(
    "driver_factory",
    [
        (
            lambda: driver.Driver(
                _dynamic_config, tests.resources.dynamic_config, adapter=DefaultAdapter()
            )
        ),
        (
            lambda: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.dynamic_config)
            .with_config(_dynamic_config)
            .with_remote_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .with_local_executor(executors.MultiThreadingExecutor(max_tasks=3))
            .build()
        ),
        (
            lambda: driver.Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.dynamic_config)
            .with_config(_dynamic_config)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_local_executor(executors.SynchronousLocalTaskExecutor())
            .build()
        ),
    ],
    ids=["basic_driver", "driver_v2_multithreading", "driver_v2_synchronous"],
)
def test_end_to_end_with_dynamic_config(driver_factory):
    dr = driver_factory()
    out = dr.execute(final_vars=list(_dynamic_config["columns_to_sum_map"].keys()))
    assert out["ab"][0] == 2
    assert out["cd"][0] == 2
    assert out["ed"][0] == 2
    assert out["ae"][0] == 2
    assert out["abcd"][0] == 4
    assert out["abcdcd"][0] == 6


class JoinBuilder(base.ResultMixin):
    @staticmethod
    def build_result(**outputs: Dict[str, Any]) -> Any:
        out = {}
        for output in outputs.values():
            out.update(output)
        return out

    def output_type(self) -> Type:
        return dict

    def input_types(self) -> List[Type]:
        return [dict]


def test_materialize_driver_call_end_to_end(tmp_path_factory):
    dr = driver.Driver({}, tests.resources.test_for_materialization)
    path_1 = tmp_path_factory.mktemp("home") / "test_materialize_driver_call_end_to_end_1.json"
    path_2 = tmp_path_factory.mktemp("home") / "test_materialize_driver_call_end_to_end_2.json"

    materialization_result, result = dr.materialize(
        to.json(id="materializer_1", dependencies=["json_to_save_1"], path=path_1),
        to.json(
            id="materializer_2",
            dependencies=["json_to_save_1", "json_to_save_2"],
            path=path_2,
            combine=JoinBuilder(),
        ),
        additional_vars=["json_to_save_1", "json_to_save_2"],
    )
    assert result == {
        "json_to_save_1": tests.resources.test_for_materialization.json_to_save_1(),
        "json_to_save_2": tests.resources.test_for_materialization.json_to_save_2(),
    }
    assert "materializer_1" in materialization_result
    assert "materializer_2" in materialization_result
    with open(path_1) as f_1, open(path_2) as f_2:
        assert json.load(f_1) == tests.resources.test_for_materialization.json_to_save_1()
        assert json.load(f_2) == {
            **tests.resources.test_for_materialization.json_to_save_2(),
            **tests.resources.test_for_materialization.json_to_save_1(),
        }


def test_materialize_and_loaders_end_to_end(tmp_path_factory):
    def processed_data(input_data: dict) -> dict:
        data = input_data.copy()
        data["processed"] = True
        return data

    path_in = tmp_path_factory.mktemp("home") / "unprocessed_data.json"
    path_out = tmp_path_factory.mktemp("home") / "processed_data.json"

    with open(path_in, "w") as f:
        json.dump({"processed": False}, f)

    mod = ad_hoc_utils.create_temporary_module(processed_data)

    dr = driver.Driver({}, mod)

    materialization_result, result = dr.materialize(
        from_.json(target="input_data", path=value(path_in)),
        to.json(
            id="materializer",
            dependencies=["processed_data"],
            path=source("output_path"),
            combine=JoinBuilder(),
        ),
        additional_vars=["processed_data"],
        inputs={"output_path": str(path_out)},
    )
    assert result["processed_data"] == {"processed": True}
    assert "materializer" in materialization_result

    with open(path_out) as f:
        assert json.load(f) == {"processed": True}


def test_materialize_and_loaders_end_to_end_without_additional_vars(tmp_path_factory):
    def processed_data(input_data: dict) -> dict:
        data = input_data.copy()
        data["processed"] = True
        return data

    path_in = tmp_path_factory.mktemp("home") / "unprocessed_data.json"
    path_out = tmp_path_factory.mktemp("home") / "processed_data.json"

    with open(path_in, "w") as f:
        json.dump({"processed": False}, f)

    mod = ad_hoc_utils.create_temporary_module(processed_data)

    dr = driver.Driver({}, mod)

    materialization_result, result = dr.materialize(
        from_.json(target="input_data", path=value(path_in)),
        to.json(
            id="materializer",
            dependencies=["processed_data"],
            path=source("output_path"),
            combine=JoinBuilder(),
        ),
        inputs={"output_path": str(path_out)},
    )
    assert "materializer" in materialization_result

    with open(path_out) as f:
        assert json.load(f) == {"processed": True}


def test_materialize_and_loaders_end_to_end_without_to(tmp_path_factory):
    def processed_data(input_data: dict) -> dict:
        data = input_data.copy()
        data["processed"] = True
        return data

    path_in = tmp_path_factory.mktemp("home") / "unprocessed_data.json"

    with open(path_in, "w") as f:
        json.dump({"processed": False}, f)

    mod = ad_hoc_utils.create_temporary_module(processed_data)

    dr = driver.Driver({}, mod)

    materialization_result, result = dr.materialize(
        from_.json(target="input_data", path=value(path_in)), additional_vars=["processed_data"]
    )
    assert result["processed_data"] == {"processed": True}
    assert "materializer" not in materialization_result


def test_no_materialize_failure(tmp_path_factory):
    def processed_data(input_data: dict) -> dict:
        data = input_data.copy()
        data["processed"] = True
        return data

    path_in = tmp_path_factory.mktemp("home") / "unprocessed_data.json"
    path_out = tmp_path_factory.mktemp("home") / "processed_data.json"

    with open(path_in, "w") as f:
        json.dump({"processed": False}, f)

    mod = ad_hoc_utils.create_temporary_module(processed_data)

    dr = driver.Driver({}, mod)

    with pytest.raises(ValueError):
        dr.materialize(
            from_.json(target="input_data", path=value(path_in)),
            inputs={"output_path": str(path_out)},
        )


def test_driver_validate_with_overrides():
    dr = (
        driver.Builder()
        .with_modules(tests.resources.overrides)
        .with_adapter(base.DefaultAdapter())
        .build()
    )
    assert dr.execute(["c"], overrides={"b": 1})["c"] == 2


def test_driver_validate_with_overrides_2():
    dr = (
        driver.Builder()
        .with_modules(tests.resources.overrides)
        .with_adapter(base.DefaultAdapter())
        .build()
    )
    assert dr.execute(["d"], overrides={"b": 1})["d"] == 3


def test_driver_validate_module_overrides():
    dr = (
        driver.Builder()
        .with_modules(tests.resources.overrides, tests.resources.overrides_from_module)
        .with_adapter(base.DefaultAdapter())
        .allow_module_overrides()
        .build()
    )
    assert dr.execute(["d"], overrides={"b": 1})["d"] == 15


def test_driver_extra_inputs_can_be_outputs():
    """Tests that we can request outputs that not in the graph, but are in the inputs."""
    dr = (
        driver.Builder()
        .with_modules(tests.resources.overrides)
        .with_adapter(base.DefaultAdapter())
        .build()
    )
    actual = dr.execute(["d", "e"], inputs={"a": 1, "e": 10})
    assert actual["d"] == 4
    assert actual["e"] == 10
    # validates validate functions
    dr.validate_execution(["d", "e"], inputs={"a": 1, "e": 10})
    dr.validate_materialization(additional_vars=["d", "e"], inputs={"a": 1, "e": 10})
    # Checks dataframe use case
    dr = (
        driver.Builder()
        .with_modules(tests.resources.example_module)
        .with_adapter(base.PandasDataFrameResult())
        .build()
    )
    actual = dr.execute(
        ["avg_3wk_spend", "e"],
        inputs={"spend": pd.Series([1, 1, 1, 1, 1]), "e": pd.Series([10, 10, 10, 10, 10])},
    )
    pd.testing.assert_frame_equal(
        actual,
        pd.DataFrame(
            {
                "avg_3wk_spend": pd.Series([math.nan, math.nan, 1.0, 1.0, 1.0], dtype=float),
                "e": [10, 10, 10, 10, 10],
            }
        ),
    )


def test_driver_v2_extra_inputs_can_be_outputs():
    """Tests that we can request outputs that not in the graph, but are in the inputs."""
    dr = (
        driver.Builder()
        .with_modules(tests.resources.overrides)
        .with_adapter(base.DefaultAdapter())
        .enable_dynamic_execution(allow_experimental_mode=True)
        .build()
    )
    actual = dr.execute(["d", "e"], inputs={"a": 1, "e": 10})
    assert actual["d"] == 4
    assert actual["e"] == 10


def test_driver_fails_on_outputs_not_in_input():
    """Tests that we fail correctly on outputs that are not in the inputs or the graph."""
    dr = (
        driver.Builder()
        .with_modules(tests.resources.overrides)
        .with_adapter(base.DefaultAdapter())
        .build()
    )
    with pytest.raises(ValueError):
        # f missing
        dr.execute(["d", "f"], inputs={"a": 1})


def test_driver_v2_fails_on_outputs_not_in_input():
    """Tests that we fail correctly on outputs that are not in the inputs or the graph."""
    dr = (
        driver.Builder()
        .with_modules(tests.resources.overrides)
        .with_adapter(base.DefaultAdapter())
        .enable_dynamic_execution(allow_experimental_mode=True)
        .build()
    )
    with pytest.raises(ValueError):
        # f missing
        dr.execute(["d", "f"], inputs={"a": 1})


def test_driver_v2_inputs_can_be_none():
    """Tests that input can be None and checks will still work."""
    dr = (
        driver.Builder()
        .with_modules(tests.resources.overrides)
        .with_adapter(base.DefaultAdapter())
        .build()
    )
    actual = dr.execute(["d"], inputs=None, overrides={"b": 1})
    assert actual["d"] == 3

    with pytest.raises(ValueError):
        # validate that None doesn't cause issues
        dr.execute(["e"], inputs=None)



import dataclasses
from os import PathLike
from typing import Any, Collection, Dict, Optional, Type, Union

try:
    import sklearn.inspection
    import sklearn.metrics
    import sklearn.model_selection
    from matplotlib import pyplot
except ImportError as e:
    raise NotImplementedError("scikit-learn is not installed.") from e


from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataSaver

display_classes = [
    "ConfusionMatrixDisplay",
    "DetCurveDisplay",
    "PrecisionRecallDisplay",
    "PredictionErrorDisplay",
    "RocCurveDisplay",
    "CalibrationDisplay",
    "DecisionBoundaryDisplay",
    "LearningCurveDisplay",
    "PartialDependenceDisplay",
    "ValidationCurveDisplay",
]
SKLEARN_PLOT_TYPES = []
for class_name in display_classes:
    # get the attribute via string from sklearn.metrics; if not found return None
    class_metrics = getattr(sklearn.metrics, class_name, None)
    class_inspection = getattr(sklearn.inspection, class_name, None)
    class_model_selection = getattr(sklearn.model_selection, class_name, None)
    if class_metrics:
        SKLEARN_PLOT_TYPES.append(class_metrics)
    if class_inspection:
        SKLEARN_PLOT_TYPES.append(class_inspection)
    if class_model_selection:
        SKLEARN_PLOT_TYPES.append(class_model_selection)

SKLEARN_PLOT_TYPES.append(pyplot.Figure)
SKLEARN_PLOT_TYPES_ANNOTATION = Union[tuple(SKLEARN_PLOT_TYPES)]


@dataclasses.dataclass
class SklearnPlotSaver(DataSaver):
    path: Union[str, PathLike]
    # kwargs
    dpi: float = 200
    format: str = "png"
    metadata: Optional[dict] = None
    bbox_inches: str = None
    pad_inches: float = 0.1
    backend: Optional[str] = None
    papertype: str = None
    transparent: bool = None
    bbox_extra_artists: Optional[list] = None
    pil_kwargs: Optional[dict] = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return SKLEARN_PLOT_TYPES

    def _get_saving_kwargs(self) -> Dict[str, Any]:
        kwargs = {}
        if self.dpi is not None:
            kwargs["dpi"] = self.dpi
        if self.format is not None:
            kwargs["format"] = self.format
        if self.metadata is not None:
            kwargs["metadata"] = self.metadata
        if self.bbox_inches is not None:
            kwargs["bbox_inches"] = self.bbox_inches
        if self.pad_inches is not None:
            kwargs["pad_inches"] = self.pad_inches
        if self.backend is not None:
            kwargs["backend"] = self.backend
        if self.papertype is not None:
            kwargs["papertype"] = self.papertype
        if self.transparent is not None:
            kwargs["transparent"] = self.transparent
        if self.bbox_extra_artists is not None:
            kwargs["bbox_extra_artists"] = self.bbox_extra_artists
        if self.pil_kwargs is not None:
            kwargs["pil_kwargs"] = self.pil_kwargs
        return kwargs

    def save_data(self, data: SKLEARN_PLOT_TYPES_ANNOTATION) -> Dict[str, Any]:
        if isinstance(data, pyplot.Figure):
            figure = data
        else:
            data.plot()
            figure = data.figure_
        figure.savefig(self.path, **self._get_saving_kwargs())
        return utils.get_file_metadata(self.path)

    @classmethod
    def name(cls) -> str:
        return "png"


def register_data_loaders():
    """Function to register the data loaders for this extension."""
    registry.register_adapter(SklearnPlotSaver)


register_data_loaders()

COLUMN_FRIENDLY_DF_TYPE = False



import dataclasses
from io import BytesIO
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Type,
    Union,
)

try:
    import polars as pl
except ImportError as e:
    raise NotImplementedError("Polars is not installed.") from e


# for polars <0.16.0 we need to determine whether type_aliases exist.
has_alias = False
if hasattr(pl, "type_aliases"):
    has_alias = True

# for polars 0.18.0 we need to check what to do.
if has_alias and hasattr(pl.type_aliases, "CsvEncoding"):
    from polars.type_aliases import CsvEncoding
else:
    CsvEncoding = Type

# import these types to make type hinting work
from polars.datatypes import DataType, DataTypeClass  # noqa: F401

from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataLoader

DATAFRAME_TYPE = pl.LazyFrame
COLUMN_TYPE = None
COLUMN_FRIENDLY_DF_TYPE = False


def register_types():
    """Function to register the types for this extension."""
    registry.register_types("polars_lazyframe", DATAFRAME_TYPE, COLUMN_TYPE)


register_types()


@dataclasses.dataclass
class PolarsScanCSVReader(DataLoader):
    """Class specifically to handle loading CSV files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_csv.html
    """

    file: Union[str, TextIO, BytesIO, Path, BinaryIO, bytes]
    # kwargs:
    has_header: bool = True
    columns: Union[Sequence[int], Sequence[str]] = None
    new_columns: Sequence[str] = None
    separator: str = ","
    comment_char: str = None
    quote_char: str = '"'
    skip_rows: int = 0
    dtypes: Union[Mapping[str, Any], Sequence[Any]] = None
    null_values: Union[str, Sequence[str], Dict[str, str]] = None
    missing_utf8_is_empty_string: bool = False
    ignore_errors: bool = False
    try_parse_dates: bool = False
    n_threads: int = None
    infer_schema_length: int = 100
    batch_size: int = 8192
    n_rows: int = None
    encoding: Union[CsvEncoding, str] = "utf8"
    low_memory: bool = False
    rechunk: bool = True
    use_pyarrow: bool = False
    storage_options: Dict[str, Any] = None
    skip_rows_after_header: int = 0
    row_count_name: str = None
    row_count_offset: int = 0
    eol_char: str = "\n"
    raise_if_empty: bool = True

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.has_header is not None:
            kwargs["has_header"] = self.has_header
        if self.columns is not None:
            kwargs["columns"] = self.columns
        if self.new_columns is not None:
            kwargs["new_columns"] = self.new_columns
        if self.separator is not None:
            kwargs["separator"] = self.separator
        if self.comment_char is not None:
            kwargs["comment_char"] = self.comment_char
        if self.quote_char is not None:
            kwargs["quote_char"] = self.quote_char
        if self.skip_rows is not None:
            kwargs["skip_rows"] = self.skip_rows
        if self.dtypes is not None:
            kwargs["dtypes"] = self.dtypes
        if self.null_values is not None:
            kwargs["null_values"] = self.null_values
        if self.missing_utf8_is_empty_string is not None:
            kwargs["missing_utf8_is_empty_string"] = self.missing_utf8_is_empty_string
        if self.ignore_errors is not None:
            kwargs["ignore_errors"] = self.ignore_errors
        if self.try_parse_dates is not None:
            kwargs["try_parse_dates"] = self.try_parse_dates
        if self.n_threads is not None:
            kwargs["n_threads"] = self.n_threads
        if self.infer_schema_length is not None:
            kwargs["infer_schema_length"] = self.infer_schema_length
        if self.n_rows is not None:
            kwargs["n_rows"] = self.n_rows
        if self.encoding is not None:
            kwargs["encoding"] = self.encoding
        if self.low_memory is not None:
            kwargs["low_memory"] = self.low_memory
        if self.rechunk is not None:
            kwargs["rechunk"] = self.rechunk
        if self.storage_options is not None:
            kwargs["storage_options"] = self.storage_options
        if self.skip_rows_after_header is not None:
            kwargs["skip_rows_after_header"] = self.skip_rows_after_header
        if self.row_count_name is not None:
            kwargs["row_count_name"] = self.row_count_name
        if self.row_count_offset is not None:
            kwargs["row_count_offset"] = self.row_count_offset
        if self.eol_char is not None:
            kwargs["eol_char"] = self.eol_char
        if self.raise_if_empty is not None:
            kwargs["raise_if_empty"] = self.raise_if_empty
        return kwargs

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.scan_csv(self.file, **self._get_loading_kwargs())

        metadata = utils.get_file_and_dataframe_metadata(self.file, df)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "csv"


@dataclasses.dataclass
class PolarsScanParquetReader(DataLoader):
    """Class specifically to handle loading parquet files with polars
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_parquet.html
    """

    file: Union[str, TextIO, BytesIO, Path, BinaryIO, bytes]
    # kwargs:
    columns: Union[List[int], List[str]] = None
    n_rows: int = None
    use_pyarrow: bool = False
    memory_map: bool = True
    storage_options: Dict[str, Any] = None
    parallel: Any = "auto"
    row_count_name: str = None
    row_count_offset: int = 0
    low_memory: bool = False
    use_statistics: bool = True
    rechunk: bool = True

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.columns is not None:
            kwargs["columns"] = self.columns
        if self.n_rows is not None:
            kwargs["n_rows"] = self.n_rows
        if self.storage_options is not None:
            kwargs["storage_options"] = self.storage_options
        if self.parallel is not None:
            kwargs["parallel"] = self.parallel
        if self.row_count_name is not None:
            kwargs["row_count_name"] = self.row_count_name
        if self.row_count_offset is not None:
            kwargs["row_count_offset"] = self.row_count_offset
        if self.low_memory is not None:
            kwargs["low_memory"] = self.low_memory
        if self.use_statistics is not None:
            kwargs["use_statistics"] = self.use_statistics
        if self.rechunk is not None:
            kwargs["rechunk"] = self.rechunk
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.scan_parquet(self.file, **self._get_loading_kwargs())
        metadata = utils.get_file_and_dataframe_metadata(self.file, df)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "parquet"


@dataclasses.dataclass
class PolarsScanFeatherReader(DataLoader):
    """
    Class specifically to handle loading Feather/Arrow IPC files with Polars.
    Should map to https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.read_ipc.html
    """

    source: Union[str, BinaryIO, BytesIO, Path, bytes]
    # kwargs:
    columns: Optional[Union[List[str], List[int]]] = None
    n_rows: Optional[int] = None
    use_pyarrow: bool = False
    memory_map: bool = True
    storage_options: Optional[Dict[str, Any]] = None
    row_count_name: Optional[str] = None
    row_count_offset: int = 0
    rechunk: bool = True

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [DATAFRAME_TYPE]

    def _get_loading_kwargs(self):
        kwargs = {}
        if self.columns is not None:
            kwargs["columns"] = self.columns
        if self.n_rows is not None:
            kwargs["n_rows"] = self.n_rows
        if self.memory_map is not None:
            kwargs["memory_map"] = self.memory_map
        if self.storage_options is not None:
            kwargs["storage_options"] = self.storage_options
        if self.row_count_name is not None:
            kwargs["row_count_name"] = self.row_count_name
        if self.row_count_offset is not None:
            kwargs["row_count_offset"] = self.row_count_offset
        if self.rechunk is not None:
            kwargs["rechunk"] = self.rechunk
        return kwargs

    def load_data(self, type_: Type) -> Tuple[DATAFRAME_TYPE, Dict[str, Any]]:
        df = pl.scan_ipc(self.source, **self._get_loading_kwargs())
        metadata = utils.get_file_metadata(self.source)
        return df, metadata

    @classmethod
    def name(cls) -> str:
        return "feather"


def register_data_loaders():
    """Function to register the data loaders for this extension."""
    for loader in [
        PolarsScanCSVReader,
        PolarsScanParquetReader,
        PolarsScanFeatherReader,
    ]:
        registry.register_adapter(loader)


register_data_loaders()



import dataclasses
from os import PathLike
from typing import (
    Any,
    BinaryIO,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

try:
    from datasets import (
        Dataset,
        DatasetDict,
        DownloadConfig,
        DownloadMode,
        Features,
        IterableDataset,
        IterableDatasetDict,
        VerificationMode,
        Version,
        load_dataset,
    )
    from datasets.formatting.formatting import LazyBatch
except ImportError as e:
    raise NotImplementedError("huggingface datasets library is not installed.") from e

try:
    import lancedb
    from lancedb import table  # noqa: F401
except ImportError:
    lancedb = None

from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataLoader, DataSaver

COLUMN_FRIENDLY_DF_TYPE = False

HF_types = (DatasetDict, Dataset, IterableDatasetDict, IterableDataset)


@dataclasses.dataclass
class HuggingFaceDSLoader(DataLoader):
    """Data loader for hugging face datasets. Uses load_data method."""

    path: str
    dataset_name: Optional[str] = None  # this can't be `name` because it clashes with `.name()`
    data_dir: Optional[str] = None
    data_files: Optional[Union[str, Sequence[str], Mapping[str, Union[str, Sequence[str]]]]] = None
    split: Optional[str] = None
    cache_dir: Optional[str] = None
    features: Optional[Features] = None
    download_config: Optional[DownloadConfig] = None
    download_mode: Optional[Union[DownloadMode, str]] = None
    verification_mode: Optional[Union[VerificationMode, str]] = None
    ignore_verifications = "deprecated"
    keep_in_memory: Optional[bool] = None
    save_infos: bool = False
    revision: Optional[Union[str, Version]] = None
    token: Optional[Union[bool, str]] = None
    use_auth_token = "deprecated"
    task = "deprecated"
    streaming: bool = False
    num_proc: Optional[int] = None
    storage_options: Optional[Dict] = None
    config_kwargs: Optional[Dict] = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return list(HF_types)

    def _get_loading_kwargs(self) -> dict:
        # Puts kwargs in a dict
        kwargs = dataclasses.asdict(self)
        # we send path separately
        del kwargs["path"]
        config_kwargs: Optional[dict] = kwargs.pop("config_kwargs", None)
        if config_kwargs:
            # add config kwargs as needed.
            kwargs.update(config_kwargs)

        # need to pass in name
        kwargs["name"] = kwargs.pop("dataset_name", None)

        return kwargs

    def load_data(self, type_: Type) -> Tuple[Union[HF_types], Dict[str, Any]]:
        """Loads the data set given the path and class values."""
        ds = load_dataset(self.path, **self._get_loading_kwargs())
        is_dataset = isinstance(ds, Dataset)
        f_meta = {"path": self.path}
        ds_meta = {"rows": ds.num_rows, "columns": ds.column_names}
        if is_dataset:
            ds_meta["size_in_bytes"] = ds.size_in_bytes
            ds_meta["features"] = ds.features.to_dict()
        return ds, {"file_metadata": f_meta, "dataset_metadata": ds_meta}

    @classmethod
    def name(cls) -> str:
        return "hf_dataset"


@dataclasses.dataclass
class HuggingFaceDSParquetSaver(DataSaver):
    """Saves a Huggingface dataset to parquet."""

    path_or_buf: Union[PathLike, BinaryIO]
    batch_size: Optional[int] = None
    parquet_writer_kwargs: Optional[dict] = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return list(HF_types)

    @classmethod
    def applies_to(cls, type_: Type[Type]) -> bool:
        return type_ in HF_types

    def _get_saving_kwargs(self) -> dict:
        # Puts kwargs in a dict
        kwargs = dataclasses.asdict(self)
        # we put path_or_buff as a positional argument
        del kwargs["path_or_buf"]
        parquet_writer_kwargs: Optional[dict] = kwargs.pop("parquet_writer_kwargs", None)
        if parquet_writer_kwargs:
            # add config kwargs as needed.
            kwargs.update(parquet_writer_kwargs)

        return kwargs

    def save_data(self, ds: Union[HF_types]) -> Dict[str, Any]:
        """Saves the data to parquet."""
        is_dataset = isinstance(ds, Dataset)
        ds.to_parquet(self.path_or_buf, **self._get_saving_kwargs())
        ds_meta = {
            "rows": ds.num_rows,
            "columns": ds.column_names,
        }
        if is_dataset:
            ds_meta.update({"size_in_bytes": ds.size_in_bytes, "features": ds.features.to_dict()})
        if isinstance(self.path_or_buf, BinaryIO):
            f_meta = {}
        else:
            f_meta = (utils.get_file_metadata(self.path_or_buf),)
        return {"file_metadata": f_meta, "dataset_metadata": ds_meta}

    @classmethod
    def name(cls) -> str:
        return "parquet"


# we do this here just in case lancedb is not installed.
if lancedb is not None:

    def _batch_write(
        dataset_batch: LazyBatch, db: lancedb.DBConnection, table_name: str, columns: str
    ) -> None:
        """Helper function to batch write to lancedb."""
        if columns is None:
            data = dataset_batch.pa_table
        else:
            data = dataset_batch.pa_table.select(columns)
        try:
            db.create_table(table_name, data)
        except (OSError, ValueError):
            tbl = db.open_table(table_name)
            tbl.add(data)
        return None

    @dataclasses.dataclass
    class HuggingFaceDSLanceDBSaver(DataSaver):
        """Data saver that saves Huggingface datasets to lancedb."""

        db_client: lancedb.DBConnection
        table_name: str
        columns_to_write: List[str] = None  # None means all.
        write_batch_size: int = 100

        @classmethod
        def applicable_types(cls) -> Collection[Type]:
            return list(HF_types)

        def save_data(self, ds: Union[HF_types]) -> Dict[str, Any]:
            """This batches writes to lancedb."""
            ds.map(
                _batch_write,
                batched=True,
                batch_size=self.write_batch_size,
                fn_kwargs={
                    "db": self.db_client,
                    "table_name": self.table_name,
                    "columns": self.columns_to_write,
                },
                desc=f"writing to lancedb table {self.table_name}",
            )
            is_dataset = isinstance(ds, Dataset)
            ds_meta = {
                "rows": ds.num_rows,
                "columns": ds.column_names,
            }
            if is_dataset:
                ds_meta.update(
                    {"size_in_bytes": ds.size_in_bytes, "features": ds.features.to_dict()}
                )
            return {"db_meta": {"table_name": self.table_name}, "dataset_metadata": ds_meta}

        @classmethod
        def name(cls) -> str:
            return "lancedb"


def register_data_loaders_savers():
    loaders = [HuggingFaceDSLoader, HuggingFaceDSParquetSaver]
    if lancedb:
        loaders.append(HuggingFaceDSLanceDBSaver)
    for loader in loaders:
        registry.register_adapter(loader)


register_data_loaders_savers()
